# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Figures of merit (FOMs) for comparison against a known ground truth."""

from __future__ import division

import numpy as np

import odl
from odl.contrib.fom.util import spherical_sum
from odl.core.discr.grid import sparse_meshgrid

__all__ = ('mean_squared_error', 'mean_absolute_error',
           'mean_value_difference', 'standard_deviation_difference',
           'range_difference', 'blurring', 'false_structures_mask', 'ssim',
           'psnr', 'haarpsi', 'noise_power_spectrum')


def mean_squared_error(data, ground_truth, mask=None,
                       normalized=False, force_lower_is_better=True):
    r"""Return mean squared L2 distance between ``data`` and ``ground_truth``.

    See also `this Wikipedia article
    <https://en.wikipedia.org/wiki/Mean_squared_error>`_.

    Parameters
    ----------
    data : `Tensor` or `array-like`
        Input data to compare to the ground truth. If not a `Tensor`, an
        unweighted tensor space will be assumed.
    ground_truth : `array-like`
        Reference to which ``data`` should be compared.
    mask : `array-like`, optional
        If given, ``data * mask`` is compared to ``ground_truth * mask``.
    normalized  : bool, optional
        If ``True``, the output values are mapped to the interval
        :math:`[0, 1]` (see `Notes` for details).
    force_lower_is_better : bool, optional
        If ``True``, it is ensured that lower values correspond to better
        matches. For the mean squared error, this is already the case, and
        the flag is only present for compatibility to other figures of merit.

    Returns
    -------
    mse : float
        FOM value, where a lower value means a better match.

    Notes
    -----
    The FOM evaluates

    .. math::
        \mathrm{MSE}(f, g) = \frac{\| f - g \|_2^2}{\| 1 \|_2^2},

    where :math:`\| 1 \|^2_2` is the volume of the domain of definition
    of the functions. For :math:`\mathbb{R}^n` type spaces, this is equal
    to the number of elements :math:`n`.

    The normalized form is

    .. math::
        \mathrm{MSE_N} = \frac{\| f - g \|_2^2}{(\| f \|_2 + \| g \|_2)^2}.

    The normalized variant takes values in :math:`[0, 1]`.
    """
    if not hasattr(data, 'space'):
        data = odl.vector(data)

    space = data.space
    ground_truth = space.element(ground_truth)

    l2norm = odl.functionals.L2Norm(space)

    if mask is not None:
        data = data * mask
        ground_truth = ground_truth * mask

    diff = data - ground_truth
    fom = l2norm(diff) ** 2

    if normalized:
        fom /= (l2norm(data) + l2norm(ground_truth)) ** 2
    else:
        fom /= l2norm(space.one()) ** 2

    # Ignore `force_lower_is_better` since that's already the case

    return fom


def mean_absolute_error(data, ground_truth, mask=None,
                        normalized=False, force_lower_is_better=True):
    r"""Return L1-distance between ``data`` and ``ground_truth``.

    See also `this Wikipedia article
    <https://en.wikipedia.org/wiki/Mean_absolute_error>`_.

    Parameters
    ----------
    data : `Tensor` or `array-like`
        Input data to compare to the ground truth. If not a `Tensor`, an
        unweighted tensor space will be assumed.
    ground_truth : `array-like`
        Reference to which ``data`` should be compared.
    mask : `array-like`, optional
        If given, ``data * mask`` is compared to ``ground_truth * mask``.
    normalized  : bool, optional
        If ``True``, the output values are mapped to the interval
        :math:`[0, 1]` (see `Notes` for details), otherwise return the
        original mean absolute error.
    force_lower_is_better : bool, optional
        If ``True``, it is ensured that lower values correspond to better
        matches. For the mean absolute error, this is already the case, and
        the flag is only present for compatibility to other figures of merit.

    Returns
    -------
    mae : float
        FOM value, where a lower value means a better match.

    Notes
    -----
    The FOM evaluates

    .. math::
        \mathrm{MAE}(f, g) = \frac{\| f - g \|_1}{\| 1 \|_1},

    where :math:`\| 1 \|_1` is the volume of the domain of definition
    of the functions. For :math:`\mathbb{R}^n` type spaces, this is equal
    to the number of elements :math:`n`.

    The normalized form is

    .. math::
        \mathrm{MAE_N}(f, g) = \frac{\| f - g \|_1}{\| f \|_1 + \| g \|_1}.

    The normalized variant takes values in :math:`[0, 1]`.
    """
    if not hasattr(data, 'space'):
        data = odl.vector(data)

    space = data.space
    ground_truth = space.element(ground_truth)

    l1_norm = odl.functionals.L1Norm(space)
    if mask is not None:
        data = data * mask
        ground_truth = ground_truth * mask

    diff = data - ground_truth
    fom = l1_norm(diff)

    if normalized:
        fom /= (l1_norm(data) + l1_norm(ground_truth))
    else:
        fom /= l1_norm(space.one())

    # Ignore `force_lower_is_better` since that's already the case

    return fom


def mean_value_difference(data, ground_truth, mask=None, normalized=False,
                          force_lower_is_better=True):
    r"""Return difference in mean value between ``data`` and ``ground_truth``.

    Parameters
    ----------
    data : `Tensor` or `array-like`
        Input data to compare to the ground truth. If not a `Tensor`, an
        unweighted tensor space will be assumed.
    ground_truth : `array-like`
        Reference to which ``data`` should be compared.
    mask : `array-like`, optional
        If given, ``data * mask`` is compared to ``ground_truth * mask``.
    normalized  : bool, optional
        Boolean flag to switch between unormalized and normalized FOM.
    force_lower_is_better : bool, optional
        If ``True``, it is ensured that lower values correspond to better
        matches. For the mean value difference, this is already the case, and
        the flag is only present for compatibility to other figures of merit.

    Returns
    -------
    mvd : float
        FOM value, where a lower value means a better match.

    Notes
    -----
    The FOM evaluates

    .. math::
         \mathrm{MVD}(f, g) =
         \Big| \overline{f} - \overline{g} \Big|,

    or, in normalized form

    .. math::
         \mathrm{MVD_N}(f, g) =
         \frac{\Big| \overline{f} - \overline{g} \Big|}
               {|\overline{f}| + |\overline{g}|}

    where :math:`\overline{f}` is the mean value of :math:`f`,

    .. math::
        \overline{f} = \frac{\langle f, 1\rangle}{\|1|_1}.

    The normalized variant takes values in :math:`[0, 1]`.
    """
    if not hasattr(data, 'space'):
        data = odl.vector(data)

    space = data.space
    ground_truth = space.element(ground_truth)

    l1_norm = odl.functionals.L1Norm(space)
    if mask is not None:
        data = data * mask
        ground_truth = ground_truth * mask

    # Volume of space
    vol = l1_norm(space.one())

    data_mean = data.inner(space.one()) / vol
    ground_truth_mean = ground_truth.inner(space.one()) / vol

    fom = np.abs(data_mean - ground_truth_mean)

    if normalized:
        fom /= (np.abs(data_mean) + np.abs(ground_truth_mean))

    # Ignore `force_lower_is_better` since that's already the case

    return fom


def standard_deviation_difference(data, ground_truth, mask=None,
                                  normalized=False,
                                  force_lower_is_better=True):
    r"""Return absolute diff in std between ``data`` and ``ground_truth``.

    Parameters
    ----------
    data : `Tensor` or `array-like`
        Input data to compare to the ground truth. If not a `Tensor`, an
        unweighted tensor space will be assumed.
    ground_truth : `array-like`
        Reference to which ``data`` should be compared.
    mask : `array-like`, optional
        If given, ``data * mask`` is compared to ``ground_truth * mask``.
    normalized  : bool, optional
        Boolean flag to switch between unormalized and normalized FOM.
    force_lower_is_better : bool, optional
        If ``True``, it is ensured that lower values correspond to better
        matches. For the standard deviation difference, this is already the
        case, and the flag is only present for compatibility to other figures
        of merit.

    Returns
    -------
    sdd : float
        FOM value, where a lower value means a better match.

    Notes
    -----
    The FOM evaluates

    .. math::
        \mathrm{SDD}(f, g) =
         \Big| \| f - \overline{f} \|_2 - \| g - \overline{g} \|_2 \Big|,

    or, in normalized form

    .. math::
        \mathrm{SDD_N}(f, g) =
         \frac{\Big| \| f - \overline{f} \|_2 -
                      \| g - \overline{g} \|_2 \Big|}
               {\| f - \overline{f} \|_2 + \| g - \overline{g} \|_2},

    where :math:`\overline{f}` is the mean value of :math:`f`,

    .. math::
        \overline{f} = \frac{\langle f, 1\rangle}{\|1|_1}.

    The normalized variant takes values in :math:`[0, 1]`.
    """
    if not hasattr(data, 'space'):
        data = odl.vector(data)

    space = data.space
    ground_truth = space.element(ground_truth)

    l1_norm = odl.functionals.L1Norm(space)
    l2_norm = odl.functionals.L2Norm(space)

    if mask is not None:
        data = data * mask
        ground_truth = ground_truth * mask

    # Volume of space
    vol = l1_norm(space.one())

    data_mean = data.inner(space.one()) / vol
    ground_truth_mean = ground_truth.inner(space.one()) / vol

    deviation_data = l2_norm(data - data_mean)
    deviation_ground_truth = l2_norm(ground_truth - ground_truth_mean)
    fom = np.abs(deviation_data - deviation_ground_truth)

    if normalized:
        denom = deviation_data + deviation_ground_truth
        if denom == 0:
            fom = 0.0
        else:
            fom /= denom

    return fom


def range_difference(data, ground_truth, mask=None, normalized=False,
                     force_lower_is_better=True):
    r"""Return dynamic range difference between ``data`` and ``ground_truth``.

    Evaluates difference in range between input (``data``) and reference
    data (``ground_truth``). Allows for normalization (``normalized``) and a
    masking of the two spaces (``mask``).

    Parameters
    ----------
    data : `odl.Tensor`
        Input data to compare to the ground truth.
    ground_truth : `odl.Tensor`
        Reference to which ``data`` should be compared.
    mask : `array-like`, optional
        Binary mask or index array to define ROI in which FOM evaluation
        is performed.
    normalized  : bool, optional
        If ``True``, normalize the FOM to lie in [0, 1].
    force_lower_is_better : bool, optional
        If ``True``, it is ensured that lower values correspond to better
        matches. For the range difference, this is already the case, and
        the flag is only present for compatibility to other figures of merit.

    Returns
    -------
    rd : float
        FOM value, where a lower value means a better match.

    Notes
    -----
    The FOM evaluates

    .. math::
        \mathrm{RD}(f, g) = \Big|
            \big(\max(f) - \min(f) \big) -
            \big(\max(g) - \min(g) \big)
            \Big|

    or, in normalized form

    .. math::
        \mathrm{RD_N}(f, g) = \frac{
            \Big|
            \big(\max(f) - \min(f) \big) -
            \big(\max(g) - \min(g) \big)
            \Big|}{
            \big(\max(f) - \min(f) \big) +
            \big(\max(g) - \min(g) \big)}

    The normalized variant takes values in :math:`[0, 1]`.
    """

    backend = data.array_backend
    ns = backend.array_namespace
    device = data.device

    data = data.asarray()
    ground_truth = ground_truth.asarray()

    if mask is not None:
        mask = data.array_backend.array_constructor(mask, dtype=bool, device=device)
        data = data[mask]
        ground_truth = ground_truth[mask]

    ptp = lambda x: float(ns.max(x) - ns.min(x))

    data_range = ptp(data)
    ground_truth_range = ptp(ground_truth)
    fom = np.abs(data_range - ground_truth_range)

    if normalized:
        denom = np.abs(data_range + ground_truth_range)
        if denom == 0:
            fom = 0.0
        else:
            fom /= denom

    return fom


def blurring(data, ground_truth, mask=None, normalized=False,
             smoothness_factor=None):
    r"""Return weighted L2 distance, emphasizing regions defined by ``mask``.

    .. note::
        If the mask argument is omitted, this FOM is equivalent to the
        mean squared error.

    Parameters
    ----------
    data : `Tensor` or `array-like`
        Input data to compare to the ground truth. If not a `Tensor`, an
        unweighted tensor space will be assumed.
    ground_truth : `array-like`
        Reference to which ``data`` should be compared.
    mask : `array-like`, optional
        Binary mask to define ROI in which FOM evaluation is performed.
    normalized  : bool, optional
        Boolean flag to switch between unormalized and normalized FOM.
    smoothness_factor : float, optional
        Positive real number. Higher value gives smoother weighting.

    Returns
    -------
    blur : float
        FOM value, where a lower value means a better match.

    See Also
    --------
    false_structures
    mean_squared_error

    Notes
    -----
    The FOM evaluates

    .. math::
        \mathrm{BLUR}(f, g) = \|\alpha (f - g) \|_2^2,

    or, in normalized form

    .. math::
        \mathrm{BLUR_N}(f, g) =
            \frac{\|\alpha(f - g)\|^2_2}
                  {\|\alpha f\|^2_2 + \|\alpha g\|^2_2}.

    The weighting function :math:`\alpha` is given as

    .. math::
        \alpha(x) = e^{-\frac{1}{k} \beta_m(x)},

    where :math:`\beta_m(x)` is the Euclidian distance transform of a
    given binary mask :math:`m`, and :math:`k` positive real number that
    controls the smoothness of the weighting function :math:`\alpha`.
    The weighting gives higher values to structures in the region of
    interest defined by the mask.

    The normalized variant takes values in :math:`[0, 1]`.
    """
    from scipy.ndimage.morphology import distance_transform_edt

    if not hasattr(data, 'space'):
        data = odl.vector(data)

    space = data.space
    ground_truth = space.element(ground_truth)

    if smoothness_factor is None:
        smoothness_factor = np.mean(data.shape) / 10

    if mask is not None:
        mask = distance_transform_edt(1 - mask)
        mask = np.exp(-mask / smoothness_factor)

    return mean_squared_error(data, ground_truth, mask, normalized)


def false_structures_mask(foreground, smoothness_factor=None):
    """Return mask emphasizing areas outside ``foreground``.

    Parameters
    ----------
    foreground : `Tensor`
        The region that should be de-emphasized. If not a `Tensor`, an
        unweighted tensor space will be assumed.
    smoothness_factor : float, optional
        Positive real number. Higher value gives smoother transition
        between foreground and its complement.

    Returns
    -------
    result : `Tensor` or `numpy.ndarray`
         Euclidean distances of elements in ``foreground``. The return value
         is a `Tensor` if ``foreground`` is one, too, otherwise a NumPy array.

    Examples
    --------
    >>> space = odl.uniform_discr(0, 1, 5)
    >>> foreground = space.element([0, 0, 1.0, 0, 0])
    >>> mask = false_structures_mask(foreground)
    >>> mask.asarray()
    array([ 0.4,  0.2,  0. ,  0.2,  0.4])

    Raises
    ------
    ValueError
        If foreground is all zero or all one, or contains values not in {0, 1}.

    Notes
    -----
    This helper function computes the Euclidean distance transform from each
    point in ``foreground.space`` to ``foreground``.

    The weighting gives higher values to structures outside the foreground
    as defined by the mask.
    """
    space = foreground.space
    has_space = True

    if space.impl != 'numpy':
        raise NotImplementedError("`false_structures_mask` currently only implemented on NumPy.")

    from scipy.ndimage.morphology import distance_transform_edt

    unique = np.unique(foreground.asarray())
    if not np.array_equiv(unique, [0., 1.]):
        raise ValueError('`foreground` is not a binary mask or has '
                         'either only true or only false values {!r}'
                         ''.format(unique))

    result = distance_transform_edt(
        1.0 - foreground.asarray(), sampling=getattr(space, 'cell_sides', 1.0)
    )
    return space.element(result)


def ssim(data, ground_truth, size=11, sigma=1.5, K1=0.01, K2=0.03,
         dynamic_range=None, normalized=False, force_lower_is_better=False):
    r"""Structural SIMilarity between ``data`` and ``ground_truth``.

    The SSIM takes value -1 for maximum dissimilarity and +1 for maximum
    similarity.

    See also `this Wikipedia article
    <https://en.wikipedia.org/wiki/Structural_similarity>`_.

    Parameters
    ----------
    data : `odl.Tensor`
        Input data to compare to the ground truth.
    ground_truth : `odl.Tensor`
        Reference to which ``data`` should be compared.
    size : odd int, optional
        Size in elements per axis of the Gaussian window that is used
        for all smoothing operations.
    sigma : positive float, optional
        Width of the Gaussian function used for smoothing.
    K1, K2 : positive float, optional
        Small constants to stabilize the result. See [Wan+2004] for details.
    dynamic_range : nonnegative float, optional
        Difference between the maximum and minimum value that the pixels
        can attain. Use 255 if pixel range is :math:`[0, 255]` and 1 if
        it is :math:`[0, 1]`. Default: `None`, obtain maximum and minimum
        from the ground truth.
    normalized  : bool, optional
        If ``True``, the output values are mapped to the interval
        :math:`[0, 1]` (see `Notes` for details), otherwise return the
        original SSIM.
    force_lower_is_better : bool, optional
        If ``True``, it is ensured that lower values correspond to better
        matches by returning the negative of the SSIM, otherwise the (possibly
        normalized) SSIM is returned. If both `normalized` and
        `force_lower_is_better` are ``True``, then the order is reversed before
        mapping the outputs, so that the latter are still in the interval
        :math:`[0, 1]`.

    Returns
    -------
    ssim : float
        FOM value, where a higher value means a better match
        if `force_lower_is_better` is ``False``.

    Notes
    -----
    The SSIM is computed on small windows and then averaged over the whole
    image. The SSIM between two windows :math:`x` and :math:`y` of size
    :math:`N \times N`

    .. math::
        SSIM(x,y) = \frac{(2\mu_x\mu_y + c_1)(2\sigma_{xy} + c_2)}
                    {(\mu_x^2 + \mu_y^2 + c_1)(\sigma_x^2 + \sigma_y^2 + c_2)}

    where:

    * :math:`\mu_x`, :math:`\mu_y` is the mean of :math:`x` and :math:`y`,
      respectively.
    * :math:`\sigma_x`, :math:`\sigma_y` is the standard deviation of
      :math:`x` and :math:`y`, respectively.
    * :math:`\sigma_{xy}` the covariance of :math:`x` and :math:`y`
    * :math:`c_1 = (k_1L)^2`, :math:`c_2 = (k_2L)^2` where :math:`L` is the
      dynamic range of the image.

    The unnormalized values are contained in the interval :math:`[-1, 1]`,
    where 1 corresponds to a perfect match. The normalized values are given by

    .. math::
        SSIM_{normalized}(x, y) = \frac{SSIM(x, y) + 1}{2}

    References
    ----------
    [Wan+2004] Wang, Z, Bovik, AC, Sheikh, HR, and Simoncelli, EP.
    *Image Quality Assessment: From Error Visibility to Structural Similarity*.
    IEEE Transactions on Image Processing, 13.4 (2004), pp 600--612.
    """

    # TODO (Justus) it would make much more sense if this function were defined
    # on a `DiscretizedSpace`, using the scale of the axes, rather than on
    # space-oblivious arrays / tensors.

    from scipy.signal import fftconvolve

    space = data.space
    backend = space.array_backend
    device = space.device

    if dynamic_range is None:
        dynamic_range = odl.max(ground_truth) - odl.min(ground_truth)

    data = data.asarray()
    ground_truth = ground_truth.asarray()

    # Compute gaussian on a `size`-sized grid in each axis
    coords = np.linspace(-(size - 1) / 2, (size - 1) / 2, size)
    grid = sparse_meshgrid(*([coords] * space.ndim))

    window = np.exp(-(sum(xi ** 2 for xi in grid) / (2.0 * sigma ** 2)))
    window /= np.sum(window)
    window = backend.array_constructor(window, device=device)

    def smoothen(img):
        """Smoothes an image by convolving with a window function."""
        return fftconvolve(window, img, mode='valid')

    C1 = (K1 * dynamic_range) ** 2
    C2 = (K2 * dynamic_range) ** 2
    mu1 = smoothen(data)
    mu2 = smoothen(ground_truth)

    mu1_sq = mu1 * mu1
    mu2_sq = mu2 * mu2
    mu1_mu2 = mu1 * mu2

    sigma1_sq = smoothen(data * data) - mu1_sq
    sigma2_sq = smoothen(ground_truth * ground_truth) - mu2_sq
    sigma12 = smoothen(data * ground_truth) - mu1_mu2

    num = (2 * mu1_mu2 + C1) * (2 * sigma12 + C2)
    denom = (mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2)
    pointwise_ssim = num / denom

    result = float(backend.array_namespace.mean(pointwise_ssim))

    if force_lower_is_better:
        result = -result

    if normalized:
        result = (result + 1.0) / 2.0

    return result


def psnr(data, ground_truth, use_zscore=False, force_lower_is_better=False):
    """Return the Peak Signal-to-Noise Ratio of ``data`` wrt ``ground_truth``.

    See also `this Wikipedia article
    <https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio>`_.

    Parameters
    ----------
    data : `Tensor`
        Input data to compare to the ground truth. If not a `Tensor`, an
        unweighted tensor space will be assumed.
    ground_truth : `Tensor`
        Reference to which ``data`` should be compared.
    use_zscore : bool
        If ``True``, normalize ``data`` and ``ground_truth`` to have zero mean
        and unit variance before comparison.
    force_lower_is_better : bool
        If ``True``, then lower value indicates better fit. In this case the
        output is negated.

    Returns
    -------
    psnr : float
        FOM value, where a higher value means a better match.

    Examples
    --------
    Compute the PSNR for two vectors:

    >>> spc = odl.rn(5)
    >>> data = spc.element([1, 1, 1, 1, 1])
    >>> ground_truth = spc.element([1, 1, 1, 1, 2])
    >>> result = psnr(data, ground_truth)
    >>> print('{:.3f}'.format(result))
    13.010

    If data == ground_truth, the result is positive infinity:

    >>> psnr(ground_truth, ground_truth)
    inf

    With ``use_zscore=True``, scaling differences and constant offsets
    are ignored:

    >>> (psnr(data, ground_truth, use_zscore=True) ==
    ...  psnr(data, 3 + 4 * ground_truth, use_zscore=True))
    True
    """
    if use_zscore:
        data = odl.core.util.zscore(data)
        ground_truth = odl.core.util.zscore(ground_truth)

    mse = mean_squared_error(data, ground_truth)
    max_true = odl.max(odl.abs(ground_truth))

    if mse == 0:
        result = np.inf
    elif max_true == 0:
        result = -np.inf
    else:
        result = 20 * np.log10(max_true) - 10 * np.log10(mse)

    if force_lower_is_better:
        return -result
    else:
        return result


def haarpsi(data, ground_truth, a=4.2, c=None):
    r"""Haar-Wavelet based perceptual similarity index FOM.

    This function evaluates the structural similarity between two images
    based on edge features along the coordinate axes, analyzed with two
    wavelet filter levels. See
    `[Rei+2016] <https://arxiv.org/abs/1607.06140>`_ and the Notes section
    for further details.

    Parameters
    ----------
    data : 2D array-like
        The image to compare to the ground truth.
    ground_truth : 2D array-like
        The true image with which to compare ``data``. It must have the
        same shape as ``data``.
    a : positive float, optional
        Parameter in the logistic function. Larger value leads to a
        steeper curve, thus lowering the threshold for an input to
        be mapped to an output close to 1. See Notes for details.
        The default value 4.2 is taken from the referenced paper.
    c : positive float, optional
        Constant determining the score of maximally dissimilar values.
        Smaller constant means higher penalty for dissimilarity.
        See `haarpsi_similarity_map` for details.
        For ``None``, the value is chosen as
        ``3 * sqrt(max(abs(ground_truth)))``.

    Returns
    -------
    haarpsi : float between 0 and 1
        The similarity score, where a higher score means a better match.
        See Notes for details.

    See Also
    --------
    haarpsi_similarity_map
    haarpsi_weight_map

    Notes
    -----
    For input images :math:`f_1, f_2`, the HaarPSI score is defined as

    .. math::
        \mathrm{HaarPSI}_{f_1, f_2} =
        l_a^{-1} \left(
        \frac{
        \sum_x \sum_{k=1}^2 \mathrm{HS}_{f_1, f_2}^{(k)}(x) \cdot
        \mathrm{W}_{f_1, f_2}^{(k)}(x)}{
        \sum_x \sum_{k=1}^2 \mathrm{W}_{f_1, f_2}^{(k)}(x)}
        \right)^2

    see `[Rei+2016] <https://arxiv.org/abs/1607.06140>`_ equation (12).

    For the definitions of the constituting functions, see

        - `haarpsi_similarity_map` for :math:`\mathrm{HS}_{f_1, f_2}^{(k)}`,
        - `haarpsi_weight_map` for :math:`\mathrm{W}_{f_1, f_2}^{(k)}`.

    References
    ----------
    [Rei+2016] Reisenhofer, R, Bosse, S, Kutyniok, G, and Wiegand, T.
    *A Haar Wavelet-Based Perceptual Similarity Index for Image Quality
    Assessment*. arXiv:1607.06140 [cs], Jul. 2016.
    """
    import scipy.special
    from odl.contrib.fom.util import haarpsi_similarity_map, haarpsi_weight_map

    if c is None:
        c = 3 * np.sqrt(np.max(np.abs(ground_truth)))

    lsim_horiz = haarpsi_similarity_map(data, ground_truth, axis=0, c=c, a=a)
    lsim_vert = haarpsi_similarity_map(data, ground_truth, axis=1, c=c, a=a)

    wmap_horiz = haarpsi_weight_map(data, ground_truth, axis=0)
    wmap_vert = haarpsi_weight_map(data, ground_truth, axis=1)

    numer = np.sum(lsim_horiz * wmap_horiz + lsim_vert * wmap_vert)
    denom = np.sum(wmap_horiz + wmap_vert)

    return (scipy.special.logit(numer / denom) / a) ** 2


def noise_power_spectrum(data, ground_truth, radial=False,
                         radial_binning_factor=2.0):
    """Return the Noise Power Spectrum (NPS).

    The NPS is given by the squared magnitude of the Fourier transform of the
    noise.

    Parameters
    ----------
    data : `DiscretizedSpaceElement` or `array-like`
        Input data to compare to the ground truth. If not a
        `DiscretizedSpaceElement`, a default space with cell size 1 will be
        assumed.
    ground_truth : `array-like`
        Reference to which ``data`` should be compared.
    radial : bool
        If ``True``, compute the radial NPS.
    radial_binning_factor : positive float, optional
        Reduce the number of radial bins by this factor. Increasing this
        number can help reducing fluctuations due to the variance of points
        that fall in a particular annulus.
        A binning factor of ``1`` corresponds to a bin size equal to
        image pixel size for images with square pixels, otherwise ::

            max(norm2(c)) / norm2(shape)

        where the maximum is taken over all corners of the image domain.

    Returns
    -------
    noise_power_spectrum : `DiscretizedSpace`-element
        The space is the Fourier space corresponding to ``space``, and
        hence the axes indicate frequency.
        If ``radial`` is ``True``, an average over concentric annuli is
        taken. The result is an element of a one-dimensional space with
        domain ``[0, rmax]``, where ``rmax`` is the radius of the smallest
        ball containing ``space.domain``. Its shape is ``(N,)`` with ::

            N = int(sqrt(sum(n ** 2 for n in image.shape)) / binning_factor)
    """
    try:
        space = data.space
        assert isinstance(space, odl.DiscretizedSpace)
    except (AttributeError, AssertionError):
        data = np.asarray(data)
        space = odl.uniform_discr(
            [0] * data.ndim, data.shape, data.shape, data.dtype
        )
        data = space.element(data)

    ft = odl.trafos.FourierTransform(space, halfcomplex=False)
    nps = np.abs(ft(data - ground_truth)).real ** 2

    if radial:
        return spherical_sum(nps, binning_factor=radial_binning_factor)
    else:
        return nps


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Cartesian products of `LinearSpace` instances."""

from __future__ import absolute_import, division, print_function

from itertools import product
from numbers import Integral, Number
import numpy as np

import warnings
from contextlib import contextmanager

from odl.core.set import LinearSpace
from odl.core.set.space import (LinearSpaceElement,
    SupportedNumOperationParadigms, NumOperationParadigmSupport)
from .weightings.weighting import (
    ArrayWeighting, ConstWeighting, CustomDist, CustomInner, CustomNorm,
    Weighting)
from odl.core.array_API_support.utils import get_array_and_backend
from odl.core.util import indent, is_real_dtype, signature_string

__all__ = ('ProductSpace',)


class ProductSpace(LinearSpace):

    """Cartesian product of `LinearSpace`'s.

    A product space is the Cartesian product ``X_1 x ... x X_n`` of
    linear spaces ``X_i``. It is itself a linear space, where the linear
    combination is defined component-wise. Inner product, norm and
    distance can also be defined in natural ways from the corresponding
    functions in the individual components.
    """

    def __init__(self, *spaces, **kwargs):
        r"""Initialize a new instance.

        Parameters
        ----------
        space1,...,spaceN : `LinearSpace` or int
            The individual spaces ("factors / parts") in the product
            space. Can also be given as ``space, n`` with ``n`` integer,
            in which case the power space ``space ** n`` is created.
        exponent : non-zero float or ``float('inf')``, optional
            Order of the product distance/norm, i.e.

            ``dist(x, y) = np.linalg.norm(x-y, ord=exponent)``

            ``norm(x) = np.linalg.norm(x, ord=exponent)``

            Values ``0 <= exponent < 1`` are currently unsupported
            due to numerical instability. See ``Notes`` for further
            information about the interpretation of the values.

            Default: 2.0

        field : `Field`, optional
            Scalar field of the resulting space.
            Default: ``spaces[0].field``

        weighting : optional
            Use weighted inner product, norm, and dist. The following
            types are supported as ``weighting``:

            ``None`` : no weighting (default)

            `Weighting` : weighting class, used directly. Such a
            class instance can be retrieved from the space by the
            `ProductSpace.weighting` property.

            `array-like` : weigh each component with one entry from the
            array. The array must be one-dimensional and have the same
            length as the number of spaces.

            float : same weighting factor in each component

        Other Parameters
        ----------------
        dist : callable, optional
            The distance function defining a metric on the space.
            It must accept two `ProductSpaceElement` arguments and
            fulfill the following mathematical conditions for any
            three space elements ``x, y, z``:

            - ``dist(x, y) >= 0``
            - ``dist(x, y) = 0``  if and only if  ``x = y``
            - ``dist(x, y) = dist(y, x)``
            - ``dist(x, y) <= dist(x, z) + dist(z, y)``

            By default, ``dist(x, y)`` is calculated as ``norm(x - y)``.

            Cannot be combined with: ``weighting, norm, inner``

        norm : callable, optional
            The norm implementation. It must accept an
            `ProductSpaceElement` argument, return a float and satisfy the
            following conditions for all space elements ``x, y`` and scalars
            ``s``:

            - ``||x|| >= 0``
            - ``||x|| = 0``  if and only if  ``x = 0``
            - ``||s * x|| = |s| * ||x||``
            - ``||x + y|| <= ||x|| + ||y||``

            By default, ``norm(x)`` is calculated as ``inner(x, x)``.

            Cannot be combined with: ``weighting, dist, inner``

        inner : callable, optional
            The inner product implementation. It must accept two
            `ProductSpaceElement` arguments, return a element from
            the field of the space (real or complex number) and
            satisfy the following conditions for all space elements
            ``x, y, z`` and scalars ``s``:

            - ``<x, y> = conj(<y, x>)``
            - ``<s*x + y, z> = s * <x, z> + <y, z>``
            - ``<x, x> = 0``  if and only if  ``x = 0``

            Cannot be combined with: ``weighting, dist, norm``

        Examples
        --------
        Product of two rn spaces

        >>> r2x3 = ProductSpace(odl.rn(2), odl.rn(3))

        Powerspace of rn space

        >>> r2x2x2 = ProductSpace(odl.rn(2), 3)

        Notes
        -----
        Inner product, norm and distance are evaluated by collecting
        the result of the corresponding operation in the individual
        components and reducing the resulting vector to a single number.
        The ``exponent`` parameter influences only this last part,
        not the computations in the individual components. We give the
        exact definitions in the following:

        Let :math:`\mathcal{X} = \mathcal{X}_1 \times \dots \times
        \mathcal{X}_d` be a product space, and
        :math:`\langle \cdot, \cdot\rangle_i`,
        :math:`\lVert \cdot \rVert_i`, :math:`d_i(\cdot, \cdot)` be
        inner products, norms and distances in the respective
        component spaces.

        **Inner product:**

        .. math::
            \langle x, y \rangle = \sum_{i=1}^d \langle x_i, y_i \rangle_i

        **Norm:**

        - :math:`p < \infty`:

        .. math::
            \lVert x\rVert =
            \left( \sum_{i=1}^d \lVert x_i \rVert_i^p \right)^{1/p}

        - :math:`p = \infty`:

        .. math::
            \lVert x\rVert = \max_i \lVert x_i \rVert_i

        **Distance:**

        - :math:`p < \infty`:

        .. math::
            d(x, y) = \left( \sum_{i=1}^d d_i(x_i, y_i)^p \right)^{1/p}

        - :math:`p = \infty`:

        .. math::
            d(x, y) = \max_i d_i(x_i, y_i)

        To implement own versions of these functions, you can use
        the following snippet to gather the vector of norms (analogously
        for inner products and distances)::

            norms = np.fromiter(
                (xi.norm() for xi in x),
                dtype=np.float64, count=len(x))

        See Also
        --------
        ProductSpaceArrayWeighting
        ProductSpaceConstWeighting
        """
        field = kwargs.pop('field', None)
        dist = kwargs.pop('dist', None)
        norm = kwargs.pop('norm', None)
        inner = kwargs.pop('inner', None)
        weighting = kwargs.pop('weighting', None)
        exponent = float(kwargs.pop('exponent', 2.0))
        if kwargs:
            raise TypeError('got unexpected keyword arguments: {}'
                            ''.format(kwargs))

        # Check validity of option combination (3 or 4 out of 4 must be None)
        if sum(x is None for x in (dist, norm, inner, weighting)) < 3:
            raise ValueError('invalid combination of options weighting, '
                             'dist, norm and inner')

        if any(x is not None for x in (dist, norm, inner)) and exponent != 2.0:
            raise ValueError('`exponent` cannot be used together with '
                             'inner, norm or dist')

        # Make a power space if the second argument is an integer.
        # For the case that the integer is 0, we already set the field here.
        if len(spaces) == 2 and isinstance(spaces[1], Integral):
            field = spaces[0].field
            spaces = [spaces[0]] * spaces[1]

        # Validate the space arguments
        if not all(isinstance(spc, LinearSpace) for spc in spaces):
            raise TypeError(
                'all arguments must be `LinearSpace` instances, or the '
                'first argument must be `LinearSpace` and the second '
                'integer; got {!r}'.format(spaces))
        if not all(spc.field == spaces[0].field for spc in spaces):
            raise ValueError('all spaces must have the same field')

        # Assign spaces and field
        self.__spaces = tuple(spaces)

        # Cache for efficiency
        self.__is_power_space = all(spc == self.spaces[0]
                                    for spc in self.spaces[1:])

        # Assing or infer field
        if field is None:
            if len(self) == 0:
                raise ValueError('no spaces provided, cannot deduce field')
            else:
                field = self.spaces[0].field

        super(ProductSpace, self).__init__(field)

        # Assign weighting
        if weighting is not None:
            if isinstance(weighting, Weighting):
                self.__weighting = weighting
            elif np.isscalar(weighting):
                self.__weighting = ProductSpaceConstWeighting(
                    weighting, exponent)
            elif weighting is None:
                # Need to wait until dist, norm and inner are handled
                pass
            else:  # last possibility: make a product space element
                arr = np.asarray(weighting)
                if arr.dtype == object:
                    raise ValueError('invalid weighting argument {}'
                                     ''.format(weighting))
                if arr.ndim == 1:
                    self.__weighting = ProductSpaceArrayWeighting(
                        arr, exponent)
                else:
                    raise ValueError('weighting array has {} dimensions, '
                                     'expected 1'.format(arr.ndim))

        elif dist is not None:
            self.__weighting = ProductSpaceCustomDist(dist)
        elif norm is not None:
            self.__weighting = ProductSpaceCustomNorm(norm)
        elif inner is not None:
            self.__weighting = ProductSpaceCustomInner(inner)
        else:  # all None -> no weighing
            self.__weighting = ProductSpaceConstWeighting(1.0, exponent)

    def __len__(self):
        """Return ``len(self)``.

        This length is the number of spaces at the top level only,
        and is equal to ``self.shape[0]``.
        """
        return len(self.spaces)

    def _elementwise_num_operation(self, operation:str
                                       , x1: LinearSpaceElement | Number
                                       , x2: None | LinearSpaceElement | Number = None
                                       , out=None
                                       , namespace=None
                                       , **kwargs ):
        """
        Internal helper function to implement the __magic_functions__ (such as __add__).

        Parameters
        ----------
        x1 : ProductSpaceElement, TensorSpaceElement, int, float, complex
            Left operand
        x2 : ProductSpaceElement, TensorSpaceElement, int, float, complex
            Right operand
        operation: str
            Attribute of the array namespace
        out : ProductSpaceElement, Optional
            ProductSpaceElement for out-of-place operations

        Returns
        -------
        ProductSpaceElement
            The result of the operation `operation` wrapped in a space with the right datatype.

        """
        if self.field is None:
            raise NotImplementedError(f"The space has no field.")
        
        if out is not None:
            if not isinstance(out, ProductSpaceElement):
                raise TypeError(f"Output argument for ProductSpace arithmetic must be a product space. {type(out)=}")
            assert len(out.parts) == len(self)

        def _dtype_adaptive_wrapper(new_parts):
            if all([xln.space == spc for xln, spc in zip(new_parts, self)]):
                return self.element(new_parts)
            else:
                # The `xl.space._elementwise_num_operation` may change the dtype, and thus the
                # part-space. For example, the `isfinite` function has boolean results.
                # In this case, the resulting product space also has the new dtype, which we
                # accomplish by creating the new space on the spot.
                new_space = ProductSpace(*[xln.space for xln in new_parts])
                return new_space.element(new_parts)

        if x2 is None:
            if out is None:
                return _dtype_adaptive_wrapper([
                    xl.space._elementwise_num_operation(operation=operation, x1=xl, namespace=namespace, **kwargs)
                    for xl in x1.parts ])
            else:
                for i, xl in enumerate(x1.parts):
                    xl.space._elementwise_num_operation(operation=operation, x1=xl, out=out.parts[i], namespace=namespace, **kwargs)
                return out

        from odl.core.operator import Operator
        if isinstance(x2, Operator):
            warnings.warn("The composition of a LinearSpaceElement and an Operator using the * operator is deprecated and will be removed in future ODL versions. Please replace * with @.")
            return x2.__rmul__(x1)

        if isinstance(x1, ProductSpaceElement) and isinstance(x2, ProductSpaceElement):
            assert len(x1.parts) == len(x2.parts)
            if out is None:
                return _dtype_adaptive_wrapper([
                    xl.space._elementwise_num_operation(operation=operation, x1=xl, x2=xr, namespace=namespace, **kwargs)
                    for xl, xr in zip(x1.parts, x2.parts) ])
            else:
                for i, xl in enumerate(x1.parts):
                    xr = x2.parts[i]
                    xl.space._elementwise_num_operation(operation=operation, x1=xl, x2=xr, out=out.parts[i], namespace=namespace, **kwargs)
                return out

        elif isinstance(x1, ProductSpaceElement):
            if out is None:
                return _dtype_adaptive_wrapper([
                    x.space._elementwise_num_operation(operation=operation, x1=x, x2=x2, namespace=namespace, **kwargs)
                    for x in x1.parts ])
            else:
                for i, x in enumerate(x1.parts):
                    x.space._elementwise_num_operation(operation=operation, x1=x, x2=x2, out=out.parts[i], namespace=namespace, **kwargs)
                return out

        elif isinstance(x2, ProductSpaceElement):
            if out is None:
                return _dtype_adaptive_wrapper([
                    x.space._elementwise_num_operation(operation=operation, x1=x1, x2=x, namespace=namespace, **kwargs)
                    for x in x2.parts ])
            else:
                for i, x in enumerate(x2.parts):
                    x.space._elementwise_num_operation(operation=operation, x1=x1, x2=x, out=out.parts[i], namespace=namespace, **kwargs)
                return out

        else:
            raise TypeError(f"At least one of the arguments to `ProductSpace._elementwise_num_operation` should be a `ProductSpaceElement`, but got {type(x1)=}, {type(x2)=}")
                
    def _element_reduction(self, operation:str
                               , x: "ProductSpaceElement"
                               , **kwargs
                               ):
        assert(x in self)
        part_results = np.array([ xp.space._element_reduction(operation, xp, **kwargs) for xp in x.parts ])
        return getattr(np, operation)(part_results).item()

    @property
    def nbytes(self):
        """Total number of bytes in memory used by an element of this space."""
        return sum(spc.nbytes for spc in self.spaces)

    @property
    def shape(self):
        """Total spaces per axis, computed recursively.

        The recursion ends at the fist level that does not have a shape.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> pspace = odl.ProductSpace(r2, r3)
        >>> pspace.shape
        (2,)
        >>> pspace2 = odl.ProductSpace(pspace, 3)
        >>> pspace2.shape
        (3, 2)

        If the space is a "pure" product space, shape recurses all the way
        into the components:

        >>> r2_2 = odl.ProductSpace(r2, 3)
        >>> r2_2.shape
        (3, 2)
        """
        if len(self) == 0:
            return ()
        elif self.is_power_space:
            try:
                sub_shape = self[0].shape
            except AttributeError:
                sub_shape = ()
        else:
            sub_shape = ()

        return (len(self),) + sub_shape

    @property
    def size(self):
        """Total number of involved spaces, computed recursively.

        The recursion ends at the fist level that does not comprise
        a *power* space, i.e., which is not made of equal spaces.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> pspace = odl.ProductSpace(r2, r3)
        >>> pspace.size
        2
        >>> pspace2 = odl.ProductSpace(pspace, 3)
        >>> pspace2.size
        6
        """
        return (0 if self.shape == () else
                int(np.prod(self.shape, dtype='int64')))

    @property
    def spaces(self):
        """A tuple containing all spaces."""
        return self.__spaces

    @property
    def is_power_space(self):
        """``True`` if all member spaces are equal."""
        return self.__is_power_space

    @property
    def exponent(self):
        """Exponent of the product space norm/dist, ``None`` for custom."""
        return self.weighting.exponent

    @property
    def weighting(self):
        """This space's weighting scheme."""
        return self.__weighting

    @property
    def is_weighted(self):
        """Return ``True`` if the space is not weighted by constant 1.0."""
        return not (
            isinstance(self.weighting, ProductSpaceConstWeighting) and
            self.weighting.const == 1.0)

    @property
    def dtype(self):
        """The data type of this space.

        This is only well defined if all subspaces have the same dtype.

        Raises
        ------
        AttributeError
            If any of the subspaces does not implement `dtype` or if the dtype
            of the subspaces does not match.
        """
        dtypes = [space.dtype for space in self.spaces]

        if all(dtype == dtypes[0] for dtype in dtypes):
            return dtypes[0]
        else:
            raise AttributeError("`dtype`'s of subspaces not equal")

    @property
    def supported_num_operation_paradigms(self) -> NumOperationParadigmSupport:
        """Whether in-place operations an out-of-place operations are supported
        depends on the subspaces. Only operations that are supported on all the
        subspaces will be supported on the product space. The style that is
        preferred on most subspaces (if any) will be chosen as preferred on the
        product space."""
        paradigms = [space.supported_num_operation_paradigms
                        for space in self.spaces]

        ip_supported = True
        ip_prefers = 0
        oop_supported = True
        oop_prefers = 0

        # Check for all of the subspaces whether they support each paradigm,
        # and count how many of them prefer each.
        for parad in paradigms:
            if parad.in_place == NumOperationParadigmSupport.NOT_SUPPORTED:
                ip_supported = False
            elif parad.in_place == NumOperationParadigmSupport.PREFERRED:
                ip_prefers += 1
            if parad.out_of_place == NumOperationParadigmSupport.NOT_SUPPORTED:
                oop_supported = False
            elif parad.out_of_place == NumOperationParadigmSupport.PREFERRED:
                oop_prefers += 1

        # Support in-place updates if all subspaces support them.
        # Prefer them if a majority of the subspaces do.
        if ip_supported:
            if ip_prefers > oop_prefers:
                in_place_support = NumOperationParadigmSupport.PREFERRED
            else:
                in_place_support = NumOperationParadigmSupport.SUPPORTED
        else:
            in_place_support = NumOperationParadigmSupport.NOT_SUPPORTED

        # Support out-of-place calculations if all subspaces support them.
        # Prefer them if a majority of the subspaces do.
        if oop_supported:
            if oop_prefers > ip_prefers:
                oo_place_support = NumOperationParadigmSupport.PREFERRED
            else:
                oo_place_support = NumOperationParadigmSupport.SUPPORTED
        else:
            oo_place_support = NumOperationParadigmSupport.NOT_SUPPORTED

        return SupportedNumOperationParadigms(
                in_place=in_place_support,
                out_of_place=oo_place_support)

    @property
    def is_real(self):
        """True if this is a space of real valued vectors."""
        return all(spc.is_real for spc in self.spaces)

    @property
    def is_complex(self):
        """True if this is a space of complex valued vectors."""
        return all(spc.is_complex for spc in self.spaces)

    @property
    def real_space(self):
        """Variant of this space with real dtype."""
        return ProductSpace(*[space.real_space for space in self.spaces])

    @property
    def complex_space(self):
        """Variant of this space with complex dtype."""
        return ProductSpace(*[space.complex_space for space in self.spaces])

    def astype(self, dtype):
        """Return a copy of this space with new ``dtype``.

        Parameters
        ----------
        dtype :
            Scalar data type of the returned space. Can be provided
            in any way the `numpy.dtype` constructor understands, e.g.
            as built-in type or as a string. Data types with non-trivial
            shapes are not allowed.

        Returns
        -------
        newspace : `ProductSpace`
            Version of this space with given data type.
        """
        if dtype is None:
            # Need to filter this out since Numpy iterprets it as 'float'
            raise ValueError('`None` is not a valid data type')

        dtype = np.dtype(dtype)
        current_dtype = getattr(self, 'dtype', object)

        if dtype == current_dtype:
            return self
        else:
            return ProductSpace(*[space.astype(dtype)
                                  for space in self.spaces])

    def element(self, inp=None, cast=True):
        """Create an element in the product space.

        Parameters
        ----------
        inp : optional
            If ``inp`` is ``None``, a new element is created from
            scratch by allocation in the spaces. If ``inp`` is
            already an element of this space, it is re-wrapped.
            Otherwise, a new element is created from the
            components by calling the ``element()`` methods
            in the component spaces.
        cast : bool, optional
            If ``True``, casting is allowed. Otherwise, a ``TypeError``
            is raised for input that is not a sequence of elements of
            the spaces that make up this product space.

        Returns
        -------
        element : `ProductSpaceElement`
            The new element

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> vec_2, vec_3 = r2.element(), r3.element()
        >>> r2x3 = ProductSpace(r2, r3)
        >>> vec_2x3 = r2x3.element()
        >>> vec_2.space == vec_2x3[0].space
        True
        >>> vec_3.space == vec_2x3[1].space
        True

        Create an element of the product space

        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> prod = ProductSpace(r2, r3)
        >>> x2 = r2.element([1, 2])
        >>> x3 = r3.element([1, 2, 3])
        >>> x = prod.element([x2, x3])
        >>> x
        ProductSpace(rn(2), rn(3)).element([
            [ 1.,  2.],
            [ 1.,  2.,  3.]
        ])
        """
        # If data is given as keyword arg, prefer it over arg list
        if inp is None:
            inp = [space.element() for space in self.spaces]

        if inp in self:
            return inp
        
        if isinstance(inp, Number):
            inp = [space.element(inp) for space in self.spaces]

        if len(inp) != len(self):
            # Here, we handle the case where the user provides an input with a single element that we will try to broadcast to all of the parts of the ProductSpace. 
            if len(inp) == 1 and cast:
                parts = [space.element(inp[0]) for space in self.spaces]
            else:                
                raise ValueError('length of `inp` {} does not match length of '
                             'space {}'.format(len(inp), len(self)))

        elif (all(isinstance(v, LinearSpaceElement) and v.space == space
                for v, space in zip(inp, self.spaces))):
            parts = list(inp)

        elif cast and len(inp) == len(self):
            # Delegate constructors
            parts = [space.element(arg)
                     for arg, space in zip(inp, self.spaces)]
            
        else:
            raise TypeError('input {!r} not a sequence of elements of the '
                            'component spaces'.format(inp))

        return self.element_type(self, parts)

    @property
    def examples(self):
        """Return examples from all sub-spaces."""
        for examples in product(*[spc.examples for spc in self.spaces]):
            name = ', '.join(name for name, _ in examples)
            element = self.element([elem for _, elem in examples])
            yield (name, element)

    def zero(self):
        """Create the zero element of the product space.

        The i-th component of the product space zero element is the
        zero element of the i-th space in the product.

        Parameters
        ----------
        None

        Returns
        -------
        zero : ProductSpaceElement
            The zero element in the product space.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> zero_2, zero_3 = r2.zero(), r3.zero()
        >>> r2x3 = ProductSpace(r2, r3)
        >>> zero_2x3 = r2x3.zero()
        >>> zero_2 == zero_2x3[0]
        True
        >>> zero_3 == zero_2x3[1]
        True
        """
        return self.element([space.zero() for space in self.spaces])

    def one(self):
        """Create the one element of the product space.

        The i-th component of the product space one element is the
        one element of the i-th space in the product.

        Parameters
        ----------
        None

        Returns
        -------
        one : ProductSpaceElement
            The one element in the product space.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> one_2, one_3 = r2.one(), r3.one()
        >>> r2x3 = ProductSpace(r2, r3)
        >>> one_2x3 = r2x3.one()
        >>> one_2 == one_2x3[0]
        True
        >>> one_3 == one_2x3[1]
        True
        """
        return self.element([space.one() for space in self.spaces])

    def _lincomb(self, a, x, b, y, out):
        """Linear combination ``out = a*x + b*y``."""
        if out is None:
            return self.element([
                space._lincomb(a, xp, b, yp, out=None)
                  for space, xp, yp in zip(self.spaces, x.parts, y.parts)])
        for space, xp, yp, outp in zip(self.spaces, x.parts, y.parts,
                                       out.parts):
            space._lincomb(a, xp, b, yp, outp)

    def _dist(self, x1, x2):
        """Distance between two elements."""
        return self.weighting.dist(x1, x2)

    def _norm(self, x):
        """Norm of an element."""
        return self.weighting.norm(x)

    def _inner(self, x1, x2):
        """Inner product of two elements."""
        return self.weighting.inner(x1, x2)

    def _multiply(self, x1, x2, out):
        """Product ``out = x1 * x2``."""
        if out is None:
            return self.element([
                spc._multiply(xp, yp, out=None)
                  for spc, xp, yp in zip(self.spaces, x1.parts, x2.parts)])
        for spc, xp, yp, outp in zip(self.spaces, x1.parts, x2.parts,
                                     out.parts):
            spc._multiply(xp, yp, outp)

    def _divide(self, x1, x2, out):
        """Quotient ``out = x1 / x2``."""
        if out is None:
            return self.element([
                spc._divide(xp, yp, out=None)
                  for spc, xp, yp in zip(self.spaces, x1.parts, x2.parts)])
        for spc, xp, yp, outp in zip(self.spaces, x1.parts, x2.parts,
                                     out.parts):
            spc._divide(xp, yp, outp)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `ProductSpace` instance, has
            the same length and the same factors. ``False`` otherwise.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> rn, rm = odl.rn(2), odl.rn(3)
        >>> r2x3, rnxm = ProductSpace(r2, r3), ProductSpace(rn, rm)
        >>> r2x3 == rnxm
        True
        >>> r3x2 = ProductSpace(r3, r2)
        >>> r2x3 == r3x2
        False
        >>> r5 = ProductSpace(*[odl.rn(1)]*5)
        >>> r2x3 == r5
        False
        >>> r5 = odl.rn(5)
        >>> r2x3 == r5
        False
        """
        if other is self:
            return True
        else:
            return (isinstance(other, ProductSpace) and
                    len(self) == len(other) and
                    self.weighting == other.weighting and
                    all(x == y for x, y in zip(self.spaces,
                                               other.spaces)))

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((type(self), self.spaces, self.weighting))

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        Examples
        --------
        Integers are used to pick components, slices to pick ranges:

        >>> r2, r3, r4 = odl.rn(2), odl.rn(3), odl.rn(4)
        >>> pspace = odl.ProductSpace(r2, r3, r4)
        >>> pspace[1]
        rn(3)
        >>> pspace[1:]
        ProductSpace(rn(3), rn(4))

        With lists, arbitrary components can be stacked together:

        >>> pspace[[0, 2, 1, 2]]
        ProductSpace(rn(2), rn(4), rn(3), rn(4))

        Tuples, i.e. multi-indices, will recursively index higher-order
        product spaces. However, remaining indices cannot be passed
        down to component spaces that are not product spaces:

        >>> pspace2 = odl.ProductSpace(pspace, 3)  # 2nd order product space
        >>> pspace2
        ProductSpace(ProductSpace(rn(2), rn(3), rn(4)), 3)
        >>> pspace2[0]
        ProductSpace(rn(2), rn(3), rn(4))
        >>> pspace2[1, 0]
        rn(2)
        >>> pspace2[:-1, 0]
        ProductSpace(rn(2), 2)
        """
        if isinstance(indices, Integral):
            return self.spaces[indices]

        elif isinstance(indices, slice):
            return ProductSpace(*self.spaces[indices], field=self.field)

        elif isinstance(indices, tuple):
            # Use tuple indexing for recursive product spaces, i.e.,
            # pspace[0, 0] == pspace[0][0]
            if not indices:
                return self
            idx = indices[0]
            if isinstance(idx, Integral):
                # Single integer in tuple, picking that space and passing
                # through the rest of the tuple. If the picked space
                # is not a product space and there are still indices left,
                # raise an error.
                space = self.spaces[idx]
                rest_indcs = indices[1:]
                if not rest_indcs:
                    return space
                elif isinstance(space, ProductSpace):
                    return space[rest_indcs]
                else:
                    raise IndexError('too many indices for recursive '
                                     'product space: remaining indices '
                                     '{}'.format(rest_indcs))
            elif isinstance(idx, slice):
                # Doing the same as with single integer with all spaces
                # in the slice, but wrapping the result into a ProductSpace.
                spaces = self.spaces[idx]
                rest_indcs = indices[1:]
                if len(spaces) == 0 and rest_indcs:
                    # Need to catch this situation since the code further
                    # down doesn't trigger an error
                    raise IndexError('too many indices for recursive '
                                     'product space: remaining indices '
                                     '{}'.format(rest_indcs))
                if not rest_indcs:
                    return ProductSpace(*spaces)
                elif all(isinstance(space, ProductSpace) for space in spaces):
                    return ProductSpace(
                        *(space[rest_indcs] for space in spaces),
                        field=self.field)
                else:
                    raise IndexError('too many indices for recursive '
                                     'product space: remaining indices '
                                     '{}'.format(rest_indcs))
            else:
                raise TypeError('index tuple can only contain'
                                'integers or slices')

        elif isinstance(indices, list):
            return ProductSpace(*[self.spaces[i] for i in indices],
                                field=self.field)

        else:
            raise TypeError('`indices` must be integer, slice, tuple or '
                            'list, got {!r}'.format(indices))

    def __str__(self):
        """Return ``str(self)``."""
        if len(self) == 0:
            return '{}'
        elif self.is_power_space:
            return '({}) ** {}'.format(self.spaces[0], len(self))
        else:
            return ' x '.join(str(space) for space in self.spaces)

    def __repr__(self):
        """Return ``repr(self)``."""
        weight_str = self.weighting.repr_part
        edgeitems = np.get_printoptions()['edgeitems']
        if len(self) == 0:
            posargs = []
            posmod = ''
            optargs = [('field', self.field, None)]
            oneline = True
        elif self.is_power_space:
            posargs = [self.spaces[0], len(self)]
            posmod = '!r'
            optargs = []
            oneline = True
        elif self.size <= 2 * edgeitems:
            posargs = self.spaces
            posmod = '!r'
            optargs = []
            argstr = ', '.join(repr(s) for s in self.spaces)
            oneline = (len(argstr + weight_str) <= 40 and
                       '\n' not in argstr + weight_str)
        else:
            posargs = (self.spaces[:edgeitems] +
                       ('...',) +
                       self.spaces[-edgeitems:])
            posmod = ['!r'] * edgeitems + ['!s'] + ['!r'] * edgeitems
            optargs = []
            oneline = False

        if oneline:
            inner_str = signature_string(posargs, optargs, sep=', ',
                                         mod=[posmod, '!r'])
            if weight_str:
                inner_str = ', '.join([inner_str, weight_str])
            return '{}({})'.format(self.__class__.__name__, inner_str)
        else:
            inner_str = signature_string(posargs, optargs, sep=',\n',
                                         mod=[posmod, '!r'])
            if weight_str:
                inner_str = ',\n'.join([inner_str, weight_str])
            return '{}(\n{}\n)'.format(self.__class__.__name__,
                                       indent(inner_str))

    @property
    def element_type(self):
        """`ProductSpaceElement`"""
        return ProductSpaceElement


class ProductSpaceElement(LinearSpaceElement):

    """Elements of a `ProductSpace`."""

    def __init__(self, space, parts):
        """Initialize a new instance."""
        super(ProductSpaceElement, self).__init__(space)
        self.__parts = tuple(parts)

    @property
    def parts(self):
        """Parts of this product space element."""
        return self.__parts

    @property
    def shape(self):
        """Number of values per axis in ``self``, computed recursively.

        The recursion ends at the fist level that does not have a shape.

        Raises
        ------
        ValueError
            If a `ProductSpace` is encountered that is not a power space.

        See Also
        --------
        ProductSpace.shape

        Examples
        --------
        >>> r4_3 = odl.ProductSpace(odl.rn(4), 3)
        >>> x = r4_3.element()
        >>> x.shape
        (3, 4)
        >>> r4_2_3 = odl.ProductSpace(r4_3, 2)
        >>> y = r4_2_3.element()
        >>> y.shape
        (2, 3, 4)
        """
        return self.space.shape

    @property
    def ndim(self):
        """Number axes in ``self``, computed recursively.

        Raises
        ------
        ValueError
            If a `ProductSpace` is encountered that is not a power space.

        See Also
        --------
        shape

        Examples
        --------
        >>> r4_3 = odl.ProductSpace(odl.rn(4), 3)
        >>> x = r4_3.element()
        >>> x.ndim
        2
        >>> r4_2_3 = odl.ProductSpace(r4_3, 2)
        >>> y = r4_2_3.element()
        >>> y.ndim
        3
        """
        return len(self.shape)

    @property
    def size(self):
        """Total number of involved spaces, computed recursively.

        See Also
        --------
        ProductSpace.size
        """
        return int(np.prod(self.shape))

    @property
    def dtype(self):
        """The data type of the space of this element."""
        return self.space.dtype

    def _assign(self, other, avoid_deep_copy):
        """Assign the values of ``other``, which is assumed to be in the
        same product space, to ``self``."""
        for tgt, src in zip(self.parts, other.parts):
            tgt.assign(src, avoid_deep_copy=avoid_deep_copy)

    def set_zero(self):
        """Set this element to zero.

        See Also
        --------
        LinearSpace.zero
        """
        for tgt in self.parts:
            tgt.set_zero()
        return self

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.space)

    @property
    def nbytes(self):
        """Total number of bytes in memory used by this element."""
        return self.space.nbytes

    def __eq__(self, other):
        """Return ``self == other``.

        Overrides the default `LinearSpace` method since it is implemented with
        the distance function, which is prone to numerical errors. This
        function checks equality per component.
        """
        if other is self:
            return True
        elif other not in self.space:
            return False
        else:
            return all(sp == op for sp, op in zip(self.parts, other.parts))

    def __getitem__(self, indices):
        """Return ``self[indices]``."""
        if isinstance(indices, Integral):
            return self.parts[indices]
        elif isinstance(indices, slice):
            return self.space[indices].element(self.parts[indices])
        elif isinstance(indices, list):
            out_parts = [self.parts[i] for i in indices]
            return self.space[indices].element(out_parts)
        elif isinstance(indices, tuple):
            if len(indices) == 0:
                return ProductSpace().element()
            elif len(indices) == 1:
                # Tuple with a single entry - we just unpack and delegate
                return self[indices[0]]
            else:
                # Tuple with multiple entries
                if isinstance(indices[0], Integral):
                    # In case the first entry is an integer, we drop the
                    # axis and return directly from `parts`
                    return self.parts[indices[0]][indices[1:]]
                else:
                    # indices[0] is a slice or list. We first retrieve the
                    # parts indexed in this axis.
                    # In any case we know that we want to keep this axis.
                    if isinstance(indices[0], list):
                        part = [self.parts[i] for i in indices[0]]
                    else:
                        part = self.parts[indices[0]]

                    if (len(indices[1:]) == 1 and
                            not all(isinstance(p, ProductSpaceElement)
                                    for p in part)):
                        # This case means we have "hit the bottom", i.e.,
                        # there are non-ProductSpaces involved. In order
                        # not to retrieve scalar values from these
                        # elements, we use a slice of size 1.
                        idx = indices[1]
                        indexed = [p[idx:idx + 1] for p in part]
                    else:
                        # Here we're still in the "product space chain",
                        # so we can use recursion to go on.
                        indexed = [p[indices[1:]] for p in part]

                    # Finally make a wrapping space for the indexed elements
                    new_space = ProductSpace(*(p.space for p in indexed))
                    return new_space.element(indexed)
        else:
            raise TypeError('bad index type {}'.format(type(indices)))

    def __setitem__(self, indices, values):
        """Implement ``self[indices] = values``."""
        # Get the parts to which we assign values
        if isinstance(indices, Integral):
            indexed_parts = (self.parts[indices],)
            values = (values,)
        elif isinstance(indices, slice):
            indexed_parts = self.parts[indices]
        elif isinstance(indices, list):
            indexed_parts = tuple(self.parts[i] for i in indices)
        elif isinstance(indices, tuple):
            if len(indices) == 0:
                return
            else:
                # We need to explicitly use __setitem__ here, otherwise
                # __getitem__ is used and assigned to, which fails if
                # a space like rn(3) is indexed at the very end.
                part = self.parts[indices[0]]
                if isinstance(part, LinearSpaceElement):
                    part.__setitem__(indices[1:], values)
                else:
                    # part is a tuple
                    for p in part:
                        p.__setitem__(indices[1:], values)
                return
        else:
            raise TypeError('bad index type {}'.format(type(indices)))

        # Do the assignment, with broadcasting if desired
        try:
            iter(values)
        except TypeError:
            # `values` is not iterable, assume it can be assigned to
            # all indexed parts
            for p in indexed_parts:
                p[:] = values
        else:
            # `values` is iterable; it could still represent a single
            # element of a power space.
            if self.space.is_power_space and values in self.space[0]:
                # Broadcast a single element across a power space
                for p in indexed_parts:
                    p[:] = values
            else:
                # Now we really have one assigned value per part
                if len(values) != len(indexed_parts):
                    raise ValueError(
                        'length of iterable `values` not equal to number of '
                        'indexed parts ({} != {})'
                        ''.format(len(values), len(indexed_parts)))
                for p, v in zip(indexed_parts, values):
                    p[:] = v

    def asarray(self, out=None, must_be_contiguous=False):
        """Extract the data of this vector as a backend-specific array.

        Only available if `is_power_space` is True.

        The ordering is such that it commutes with indexing::

            self[ind].asarray() == self.asarray()[ind]

        Parameters
        ----------
        out : Arraylike, optional
            Array in which the result should be written in-place.
            Has to be contiguous and of the correct backend,
            dtype and shape.

        Raises
        ------
        ValueError
            If `is_power_space` is false.

        Examples
        --------
        >>> spc = odl.ProductSpace(odl.rn(3), 2)
        >>> x = spc.element([[ 1.,  2.,  3.],
        ...                  [ 4.,  5.,  6.]])
        >>> x.asarray()
        array([[ 1.,  2.,  3.],
               [ 4.,  5.,  6.]])
        """
        if not self.space.is_power_space:
            raise ValueError('cannot use `asarray` if `space.is_power_space` '
                             'is `False`')
        else:
            representative_array, representative_backend = get_array_and_backend(self.parts[0])

            if out is None:
                # We are assuming that `empty` always produces a contiguous array,
                # so no need to ensure it separately.
                out = representative_backend.array_namespace.empty(
                         shape=self.shape,
                         dtype=self.dtype,
                         device=representative_array.device)

            out[0] = representative_array

            for i in range(1, len(self)):
                self.parts[i].asarray(out = out[i])

            return out

    @contextmanager
    def writable_array(self, must_be_contiguous: bool =False):
        arr = None
        try:
            arr = self.asarray(must_be_contiguous=must_be_contiguous)
            yield arr
        finally:
            if arr is not None:
                for i in range(1, len(self)):
                    self.parts[i]._assign(self.parts[i].space.element(arr[i]))
    @property
    def real(self):
        """Real part of the element.

        The real part can also be set using ``x.real = other``, where ``other``
        is array-like or scalar.

        Examples
        --------
        >>> space = odl.ProductSpace(odl.cn(3), odl.cn(2))
        >>> x = space.element([[1 + 1j, 2, 3 - 3j],
        ...                    [-1 + 2j, -2 - 3j]])
        >>> x.real
        ProductSpace(rn(3), rn(2)).element([
            [ 1.,  2.,  3.],
            [-1., -2.]
        ])

        The real part can also be set using different array-like types:

        >>> x.real = space.real_space.zero()
        >>> x
        ProductSpace(cn(3), cn(2)).element([
            [ 0.+1.j,  0.+0.j,  0.-3.j],
            [ 0.+2.j,  0.-3.j]
        ])

        >>> x.real = 1.0
        >>> x
        ProductSpace(cn(3), cn(2)).element([
            [ 1.+1.j,  1.+0.j,  1.-3.j],
            [ 1.+2.j,  1.-3.j]
        ])

        >>> x.real = [[2, 3, 4], [5, 6]]
        >>> x
        ProductSpace(cn(3), cn(2)).element([
            [ 2.+1.j,  3.+0.j,  4.-3.j],
            [ 5.+2.j,  6.-3.j]
        ])
        """
        real_part = [part.real for part in self.parts]
        return self.space.real_space.element(real_part)

    @real.setter
    def real(self, newreal):
        """Setter for the real part.

        This method is invoked by ``x.real = other``.

        Parameters
        ----------
        newreal : array-like or scalar
            Values to be assigned to the real part of this element.
        """
        try:
            iter(newreal)
        except TypeError:
            # `newreal` is not iterable, assume it can be assigned to
            # all indexed parts
            for part in self.parts:
                part.real = newreal
            return

        if self.space.is_power_space:
            try:
                # Set same value in all parts
                for part in self.parts:
                    part.real = newreal
            except (AttributeError, ValueError, TypeError):
                # Iterate over all parts and set them separately
                for part, new_re in zip(self.parts, newreal):
                    part.real = new_re
                pass
        elif len(newreal) == len(self):
            for part, new_re in zip(self.parts, newreal):
                part.real = new_re
        else:
            raise ValueError(
                'dimensions of the new real part does not match the space, '
                'got element {} to set real part of {}'.format(newreal, self))

    @property
    def imag(self):
        """Imaginary part of the element.

        The imaginary part can also be set using ``x.imag = other``, where
        ``other`` is array-like or scalar.


        Examples
        --------
        >>> space = odl.ProductSpace(odl.cn(3), odl.cn(2))
        >>> x = space.element([[1 + 1j, 2, 3 - 3j],
        ...                    [-1 + 2j, -2 - 3j]])
        >>> x.imag
        ProductSpace(rn(3), rn(2)).element([
            [ 1.,  0., -3.],
            [ 2., -3.]
        ])

        The imaginary part can also be set using different array-like types:

        >>> x.imag = space.real_space.zero()
        >>> x
        ProductSpace(cn(3), cn(2)).element([
            [ 1.+0.j,  2.+0.j,  3.+0.j],
            [-1.+0.j, -2.+0.j]
        ])

        >>> x.imag = 1.0
        >>> x
        ProductSpace(cn(3), cn(2)).element([
            [ 1.+1.j,  2.+1.j,  3.+1.j],
            [-1.+1.j, -2.+1.j]
        ])

        >>> x.imag = [[2, 3, 4], [5, 6]]
        >>> x
        ProductSpace(cn(3), cn(2)).element([
            [ 1.+2.j,  2.+3.j,  3.+4.j],
            [-1.+5.j, -2.+6.j]
        ])
        """
        imag_part = [part.imag for part in self.parts]
        return self.space.real_space.element(imag_part)

    @imag.setter
    def imag(self, newimag):
        """Setter for the imaginary part.

        This method is invoked by ``x.imag = other``.

        Parameters
        ----------
        newimag : array-like or scalar
            Values to be assigned to the imaginary part of this element.
        """
        try:
            iter(newimag)
        except TypeError:
            # `newimag` is not iterable, assume it can be assigned to
            # all indexed parts
            for part in self.parts:
                part.imag = newimag
            return

        if self.space.is_power_space:
            try:
                # Set same value in all parts
                for part in self.parts:
                    part.imag = newimag
            except (AttributeError, ValueError, TypeError):
                # Iterate over all parts and set them separately
                for part, new_im in zip(self.parts, newimag):
                    part.imag = new_im
                pass
        elif len(newimag) == len(self):
            for part, new_im in zip(self.parts, newimag):
                part.imag = new_im
        else:
            raise ValueError(
                'dimensions of the new imaginary part does not match the '
                'space, got element {} to set real part of {}}'
                ''.format(newimag, self))

    def conj(self):
        """Complex conjugate of the element."""
        complex_conj = [part.conj() for part in self.parts]
        return self.space.element(complex_conj)

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> from odl import rn  # need to import rn into namespace
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> r2x3 = ProductSpace(r2, r3)
        >>> x = r2x3.element([[1, 2], [3, 4, 5]])
        >>> eval(repr(x)) == x
        True

        The result is readable:

        >>> x
        ProductSpace(rn(2), rn(3)).element([
            [ 1.,  2.],
            [ 3.,  4.,  5.]
        ])

        Nestled spaces work as well:

        >>> X = ProductSpace(r2x3, r2x3)
        >>> x = X.element([[[1, 2], [3, 4, 5]],[[1, 2], [3, 4, 5]]])
        >>> eval(repr(x)) == x
        True
        >>> x
        ProductSpace(ProductSpace(rn(2), rn(3)), 2).element([
            [
                [ 1.,  2.],
                [ 3.,  4.,  5.]
            ],
            [
                [ 1.,  2.],
                [ 3.,  4.,  5.]
            ]
        ])
        """
        inner_str = '[\n'
        if len(self) < 5:
            inner_str += ',\n'.join('{}'.format(
                _indent(_strip_space(part))) for part in self.parts)
        else:
            inner_str += ',\n'.join('{}'.format(
                _indent(_strip_space(part))) for part in self.parts[:3])
            inner_str += ',\n    ...\n'
            inner_str += ',\n'.join('{}'.format(
                _indent(_strip_space(part))) for part in self.parts[-1:])

        inner_str += '\n]'

        return '{!r}.element({})'.format(self.space, inner_str)

    def show(self, title=None, indices=None, **kwargs):
        """Display the parts of this product space element graphically.

        Parameters
        ----------
        title : string, optional
            Title of the figures

        indices : int, slice, tuple or list, optional
            Display parts of ``self`` in the way described in the following.

            A single list of integers selects the corresponding parts
            of this vector.

            For other tuples or lists, the first entry indexes the parts of
            this vector, and the remaining entries (if any) are used to
            slice into the parts. Handling those remaining indices is
            up to the ``show`` methods of the parts to be displayed.

            The types of the first entry trigger the following behaviors:

                - ``int``: take the part corresponding to this index
                - ``slice``: take a subset of the parts
                - ``None``: equivalent to ``slice(None)``, i.e., everything

            Typical use cases are displaying of selected parts, which can
            be achieved with a list, e.g., ``indices=[0, 2]`` for parts
            0 and 2, and plotting of all parts sliced in a certain way,
            e.g., ``indices=[None, 20, None]`` for showing all parts
            sliced with indices ``[20, None]``.

            A single ``int``, ``slice``, ``list`` or ``None`` object
            indexes the parts only, i.e., is treated roughly as
            ``(indices, Ellipsis)``. In particular, for ``None``, all
            parts are shown with default slicing.

        in_figs : sequence of `matplotlib.figure.Figure`, optional
            Update these figures instead of creating new ones. Typically
            the return value of an earlier call to ``show`` is used
            for this parameter.

        kwargs
            Additional arguments passed on to the ``show`` methods of
            the parts.

        Returns
        -------
        figs : tuple of `matplotlib.figure.Figure`
            The resulting figures. In an interactive shell, they are
            automatically displayed.

        See Also
        --------
        odl.core.discr.discr_space.DiscretizedSpaceElement.show :
            Display of a discretized function
        odl.core.space.base_tensors.Tensor.show :
            Display of sequence type data
        odl.core.util.graphics.show_discrete_data :
            Underlying implementation
        """
        if title is None:
            title = 'ProductSpaceElement'

        if indices is None:
            if len(self) < 5:
                indices = list(range(len(self)))
            else:
                indices = list(np.linspace(0, len(self) - 1, 4, dtype=int))
        else:
            if (isinstance(indices, tuple) or
                    (isinstance(indices, list) and
                     not all(isinstance(idx, Integral) for idx in indices))):
                # Tuples or lists containing non-integers index by axis.
                # We use the first index for the current pspace and pass
                # on the rest.
                indices, kwargs['indices'] = indices[0], indices[1:]

            # Support `indices=[None, 0, None]` like syntax (`indices` is
            # the first entry as of now in that case)
            if indices is None:
                indices = slice(None)

            if isinstance(indices, slice):
                indices = list(range(*indices.indices(len(self))))
            elif isinstance(indices, Integral):
                indices = [indices]
            else:
                # Use `indices` as-is
                pass

        in_figs = kwargs.pop('fig', None)
        in_figs = [None] * len(indices) if in_figs is None else in_figs

        figs = []
        parts = self[indices]
        if len(parts) == 0:
            return ()
        elif len(parts) == 1:
            # Don't extend the title if there is only one plot
            fig = parts[0].show(title=title, fig=in_figs[0], **kwargs)
            figs.append(fig)
        else:
            # Extend titles by indexed part to make them distinguishable
            for i, part, fig in zip(indices, parts, in_figs):
                fig = part.show(title='{}. Part {}'.format(title, i), fig=fig,
                                **kwargs)
                figs.append(fig)

        return tuple(figs)

class ProductSpaceArrayWeighting(ArrayWeighting):

    """Array weighting for `ProductSpace`.

    This class defines a weighting that has a different value for
    each index defined in a given space.
    See ``Notes`` for mathematical details.
    """

    def __init__(self, array, exponent=2.0):
        r"""Initialize a new instance.

        Parameters
        ----------
        array : 1-dim. `array-like`
            Weighting array of the inner product.
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, no inner
            product is defined.

        Notes
        -----
        - For exponent 2.0, a new weighted inner product with array
          :math:`w` is defined as

          .. math::
              \langle x, y \rangle_w = \langle w \odot x, y \rangle

          with component-wise multiplication :math:`w \odot x`. For other
          exponents, only ``norm`` and ``dist`` are defined. In the case
          of exponent ``inf``, the weighted norm is

          .. math::
              \|x\|_{w,\infty} = \|w \odot x\|_\infty,

          otherwise it is

          .. math::
              \|x\|_{w,p} = \|w^{1/p} \odot x\|_p.

        - Note that this definition does **not** fulfill the limit property
          in :math:`p`, i.e.,

          .. math::
              \|x\|_{w,p} \not\to \|x\|_{w,\infty}
              \quad\text{for } p \to \infty

          unless :math:`w = (1,...,1)`. The reason for this choice
          is that the alternative with the limit property consists in
          ignoring the weights altogether.

        - The array may only have positive entries, otherwise it does not
          define an inner product or norm, respectively. This is not checked
          during initialization.
        """
        super(ProductSpaceArrayWeighting, self).__init__(
            array, impl=None, device=None, exponent=exponent)

    def inner(self, x1, x2):
        """Calculate the array-weighted inner product of two elements.

        Parameters
        ----------
        x1, x2 : `ProductSpaceElement`
            Elements whose inner product is calculated.

        Returns
        -------
        inner : float or complex
            The inner product of the two provided elements.
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))

        inners = np.fromiter(
            (x1i.inner(x2i) for x1i, x2i in zip(x1, x2)),
            dtype=x1[0].space.dtype_identifier, count=len(x1))

        inner = np.dot(inners, self.array)
        if is_real_dtype(x1[0].dtype):
            return float(inner)
        else:
            return complex(inner)

    def norm(self, x):
        """Calculate the array-weighted norm of an element.

        Parameters
        ----------
        x : `ProductSpaceElement`
            Element whose norm is calculated.

        Returns
        -------
        norm : float
            The norm of the provided element.
        """
        if self.exponent == 2.0:
            norm_squared = self.inner(x, x).real  # TODO: optimize?!
            return np.sqrt(norm_squared)
        else:
            norms = np.fromiter(
                (xi.norm() for xi in x), dtype=np.float64, count=len(x))
            if self.exponent in (1.0, float('inf')):
                norms *= self.array
            else:
                norms *= self.array ** (1.0 / self.exponent)

            return float(np.linalg.norm(norms, ord=self.exponent))


class ProductSpaceConstWeighting(ConstWeighting):

    """Constant weighting for `ProductSpace`.

    """

    def __init__(self, constant, exponent=2.0):
        r"""Initialize a new instance.

        Parameters
        ----------
        constant : positive float
            Weighting constant of the inner product
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, no inner
            product is defined.

        Notes
        -----
        - For exponent 2.0, a new weighted inner product with constant
          :math:`c` is defined as

          .. math::
            \langle x, y \rangle_c = c\, \langle x, y \rangle.

          For other exponents, only ``norm`` and ```dist`` are defined.
          In the case of exponent ``inf``, the weighted norm is

          .. math::
              \|x\|_{c,\infty} = c\, \|x\|_\infty,

          otherwise it is

          .. math::
              \|x\|_{c,p} = c^{1/p} \, \|x\|_p.

        - Note that this definition does **not** fulfill the limit property
          in :math:`p`, i.e.,

          .. math::
              \|x\|_{c,p} \not\to \|x\|_{c,\infty}
              \quad \text{for } p \to \infty

          unless :math:`c = 1`. The reason for this choice
          is that the alternative with the limit property consists in
          ignoring the weight altogether.

        - The constant must be positive, otherwise it does not define an
          inner product or norm, respectively.
        """
        super(ProductSpaceConstWeighting, self).__init__(
            constant, impl=None, device=None, exponent=exponent)

    def inner(self, x1, x2):
        """Calculate the constant-weighted inner product of two elements.

        Parameters
        ----------
        x1, x2 : `ProductSpaceElement`
            Elements whose inner product is calculated.

        Returns
        -------
        inner : float or complex
            The inner product of the two provided elements.
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))

        accumulator = 0.0
        # Manual loop, to avoid having to select a universally-applicable dtype
        for x1i, x2i in zip(x1, x2):
            accumulator = accumulator + x1i.inner(x2i)

        result = self.const * accumulator
        return x1.space.field.element(result)

    def norm(self, x):
        """Calculate the constant-weighted norm of an element.

        Parameters
        ----------
        x1 : `ProductSpaceElement`
            Element whose norm is calculated.

        Returns
        -------
        norm : float
            The norm of the element.
        """
        if self.exponent == 2.0:
            norm_squared = self.inner(x, x).real  # TODO: optimize?!
            return np.sqrt(norm_squared)
        else:
            norms = np.fromiter(
                (xi.norm() for xi in x), dtype=np.float64, count=len(x))

            if self.exponent in (1.0, float('inf')):
                return (self.const *
                        float(np.linalg.norm(norms, ord=self.exponent)))
            else:
                return (self.const ** (1 / self.exponent) *
                        float(np.linalg.norm(norms, ord=self.exponent)))

    def dist(self, x1, x2):
        """Calculate the constant-weighted distance between two elements.

        Parameters
        ----------
        x1, x2 : `ProductSpaceElement`
            Elements whose mutual distance is calculated.

        Returns
        -------
        dist : float
            The distance between the elements.
        """
        dnorms = np.fromiter(
            ((x1i - x2i).norm() for x1i, x2i in zip(x1, x2)),
            dtype=np.float64, count=len(x1))

        if self.exponent == float('inf'):
            return self.const * np.linalg.norm(dnorms, ord=self.exponent)
        else:
            return (self.const ** (1 / self.exponent) *
                    np.linalg.norm(dnorms, ord=self.exponent))


class ProductSpaceCustomInner(CustomInner):

    """Class for handling a user-specified inner products."""

    def __init__(self, inner):
        """Initialize a new instance.

        Parameters
        ----------
        inner : callable
            The inner product implementation. It must accept two
            `ProductSpaceElement` arguments, return a element from
            the field of the space (real or complex number) and
            satisfy the following conditions for all space elements
            ``x, y, z`` and scalars ``s``:

            - ``<x, y> = conj(<y, x>)``
            - ``<s*x + y, z> = s * <x, z> + <y, z>``
            - ``<x, x> = 0``  if and only if  ``x = 0``
        """
        super(ProductSpaceCustomInner, self).__init__(
            impl=None, inner=inner, device=None)


class ProductSpaceCustomNorm(CustomNorm):

    """Class for handling a user-specified norm on `ProductSpace`.

    Note that this removes ``inner``.
    """

    def __init__(self, norm):
        """Initialize a new instance.

        Parameters
        ----------
        norm : callable
            The norm implementation. It must accept a
            `ProductSpaceElement` argument, return a float and satisfy
            the following conditions for all space elements
            ``x, y`` and scalars ``s``:

            - ``||x|| >= 0``
            - ``||x|| = 0``  if and only if  ``x = 0``
            - ``||s * x|| = |s| * ||x||``
            - ``||x + y|| <= ||x|| + ||y||``
        """
        super(ProductSpaceCustomNorm, self).__init__(norm, impl=None, device=None)


class ProductSpaceCustomDist(CustomDist):

    """Class for handling a user-specified distance on `ProductSpace`.

    Note that this removes ``inner`` and ``norm``.
    """

    def __init__(self, dist):
        """Initialize a new instance.

        Parameters
        ----------
        dist : callable
            The distance function defining a metric on
            `ProductSpace`. It must accept two `ProductSpaceElement`
            arguments and fulfill the following mathematical conditions
            for any three space elements ``x, y, z``:

            - ``dist(x, y) >= 0``
            - ``dist(x, y) = 0``  if and only if  ``x = y``
            - ``dist(x, y) = dist(y, x)``
            - ``dist(x, y) <= dist(x, z) + dist(z, y)``
        """
        super(ProductSpaceCustomDist, self).__init__(dist, impl=None, device=None)


def _strip_space(x):
    """Strip the SPACE.element( ... ) part from a repr."""
    r = repr(x)
    space_repr = '{!r}.element('.format(x.space)
    if r.startswith(space_repr) and r.endswith(')'):
        r = r[len(space_repr):-1]
    return r


def _indent(x):
    """Indent a string by 4 characters."""
    lines = x.splitlines()
    for i, line in enumerate(lines):
        lines[i] = '    ' + line
    return '\n'.join(lines)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division

import numpy as np
import pytest

import odl
from odl.core.util.normalize import (
    normalized_axes_tuple, normalized_scalar_param_list)
from odl.core.util.testutils import simple_fixture

# --- pytest fixtures --- #


length = simple_fixture('length', [1, 2])


single_conv_params = [(-1.0, float),
                      (2, float),
                      ('10', float),
                      (np.array(2.3, dtype=float), float),
                      (np.array(2, dtype=int), float),
                      (-1, int),
                      ('10', int),
                      (np.array(2, dtype=int), int),
                      (1, bool),
                      (False, bool)]

single_conv_ids = [' input = {0[0]}, conv = {0[1]} '.format(p)
                   for p in single_conv_params]


@pytest.fixture(scope='module', ids=single_conv_ids, params=single_conv_params)
def single_conv(request):
    return request.param


lengts = [1, 2, 5]

seq_conv_params = [([-1.0], float),
                   ((2,), float),
                   (['10'], float),
                   ([np.array(2.3, dtype=float)], float),
                   ([np.array(2, dtype=int)], float),
                   ((-1,), int),
                   (['10'], int),
                   ((np.array(2, dtype=int),), int),
                   ([1], bool),
                   ((False,), bool)]


seq_conv_ids = [' input = {0[0]}, conv = {0[1]} '.format(p)
                for p in seq_conv_params]


@pytest.fixture(scope='module', ids=seq_conv_ids, params=seq_conv_params)
def seq_conv(request):
    return request.param


# For ndim = 3
axes_conv_params = [(0, (0,)),
                    (-1, (2,)),
                    ((1,), (1,)),
                    ([-1], (2,)),
                    ((1, 2, 0), (1, 2, 0)),
                    ((2, 1, -3), (2, 1, 0)),
                    ([0, 1], (0, 1)),
                    (np.arange(2), (0, 1))]
axes_conv_ids = [' axes={0[0]}, conv={0[1]} '.format(axis)
                 for axis in axes_conv_params]


@pytest.fixture(scope='module', ids=axes_conv_ids, params=axes_conv_params)
def axes_conv(request):
    return request.param


# --- normalized_scalar_param_list --- #


def test_normalized_scalar_param_list_single_val(length, single_conv):

    value, conversion = single_conv

    expected_noconv = [value] * length
    norm_param_noconv = normalized_scalar_param_list(value, length)
    assert expected_noconv == norm_param_noconv

    expected_conv = [conversion(value)] * length
    norm_param_conv = normalized_scalar_param_list(
        value, length, param_conv=conversion)
    assert expected_conv == norm_param_conv


def test_normalized_scalar_param_list_sequence(length, seq_conv):

    value, conversion = seq_conv
    value = value * length

    expected_noconv = list(value)
    norm_param_noconv = normalized_scalar_param_list(value, length)
    assert expected_noconv == norm_param_noconv

    expected_conv = [conversion(v) for v in value]
    norm_param_conv = normalized_scalar_param_list(
        value, length, param_conv=conversion)
    assert expected_conv == norm_param_conv


def test_normalized_scalar_param_list_with_none():

    param1 = [1, None, 0]

    def conv_int_none(x):
        if x is None:
            return 0
        else:
            return int(x)

    norm_param_noconv = normalized_scalar_param_list(param1, length=3)
    assert norm_param_noconv == param1

    norm_param_conv1 = normalized_scalar_param_list(
        param1, length=3, param_conv=conv_int_none, keep_none=True)
    assert norm_param_conv1 == param1

    norm_param_conv2 = normalized_scalar_param_list(
        param1, length=3, param_conv=conv_int_none, keep_none=False)
    assert norm_param_conv2 == [1, 0, 0]

    norm_param_noconv = normalized_scalar_param_list(None, length=3)
    assert norm_param_noconv == [None] * 3

    norm_param_conv1 = normalized_scalar_param_list(
        None, length=3, param_conv=conv_int_none, keep_none=True)
    assert norm_param_conv1 == [None] * 3

    norm_param_conv2 = normalized_scalar_param_list(
        None, length=3, param_conv=conv_int_none, keep_none=False)
    assert norm_param_conv2 == [0] * 3


def test_normalized_scalar_param_list_error():

    # Wrong length
    with pytest.raises(ValueError):
        normalized_scalar_param_list([1, 2], length=3)


# --- normalized_axes_tuple

def test_normalized_axes_tuple(axes_conv):
    """Test if all valid sequences are converted correctly."""
    axes, conversion = axes_conv
    assert normalized_axes_tuple(axes, ndim=3) == conversion


def test_normalized_axes_tuple_raise():
    """Test if errors are raised for invalid input."""

    with pytest.raises(TypeError):
        normalized_axes_tuple(1.5, ndim=3)  # float

    with pytest.raises(TypeError):
        normalized_axes_tuple(None, ndim=3)  # garbage

    with pytest.raises(ValueError):
        normalized_axes_tuple((0, 1.5), ndim=3)  # sequence containing float

    with pytest.raises(TypeError):
        normalized_axes_tuple((0, None), ndim=3)  # sequence containing garbge

    with pytest.raises(ValueError):
        normalized_axes_tuple((0, 0, 1), ndim=3)  # duplicate

    with pytest.raises(ValueError):
        normalized_axes_tuple((0,), ndim=0)  # nonpositive ndim

    with pytest.raises(ValueError):
        normalized_axes_tuple((0, 1), ndim=1)  # axis out of range

    with pytest.raises(ValueError):
        normalized_axes_tuple(-3, ndim=2)  # axis out of range

    with pytest.raises(ValueError):
        normalized_axes_tuple((0, 2), ndim=2)  # axis out of range


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the Functional class."""

from __future__ import division
import numpy as np
import pytest

import odl
from odl.core.operator import OpTypeError
from odl.core.util.testutils import (
    all_almost_equal, dtype_ndigits, dtype_tol, noise_element, simple_fixture)
from odl.functionals.default_functionals import (
    KullbackLeiblerConvexConj)
from odl.solvers.nonsmooth.proximal_operators import _numerical_epsilon


# TODO: maybe add tests for if translations etc. belongs to the wrong space.
# These tests don't work as intended now, since casting is possible between
# spaces with the same number of discretization points.


# --- pytest fixtures --- #


scalar = simple_fixture('scalar', [0.01, 2.7, 10, -2, -0.2, -7.1, 0])
sigma = simple_fixture('sigma', [0.001, 2.7, np.array(0.5), 10])

space_params = ['r10', 'uniform_discr', 'power_space_unif_discr']
space_ids = [' space={} '.format(p) for p in space_params]


@pytest.fixture(scope="module", ids=space_ids, params=space_params)
def space(request, odl_impl_device_pairs):
    name = request.param.strip()
    impl, device = odl_impl_device_pairs

    if name == 'r10':
        return odl.rn(10, impl=impl, device=device)
    elif name == 'uniform_discr':
        # Discretization parameters
        return odl.uniform_discr(0, 1, 7, impl=impl, device=device)
    elif name == 'power_space_unif_discr':
        # Discretization parameters
        space = odl.uniform_discr(0, 1, 7, impl=impl, device=device)
        return odl.ProductSpace(space, 2)


func_params = ['l1 ', 'l2', 'l2^2', 'constant', 'zero', 'ind_unit_ball_1',
               'ind_unit_ball_2', 'ind_unit_ball_pi', 'ind_unit_ball_inf',
               'product', 'quotient', 'kl', 'kl_cc', 'kl_cross_ent',
               'kl_cc_cross_ent', 'huber', 'groupl1', 'bregman_l2squared',
               'bregman_l1', 'indicator_simplex', 'indicator_sum_constraint']

func_ids = [" functional='{}' ".format(p) for p in func_params]

FUNCTIONALS_WITHOUT_DERIVATIVE = (
    odl.functionals.IndicatorLpUnitBall,
    odl.functionals.IndicatorSimplex,
    odl.functionals.IndicatorSumConstraint)


@pytest.fixture(scope="module", ids=func_ids, params=func_params)
def functional(request, space):
    name = request.param.strip()

    if name == 'l1':
        func = odl.functionals.L1Norm(space)
    elif name == 'l2':
        func = odl.functionals.L2Norm(space)
    elif name == 'l2^2':
        func = odl.functionals.L2NormSquared(space)
    elif name == 'constant':
        func = odl.functionals.ConstantFunctional(space, 2)
    elif name == 'zero':
        func = odl.functionals.ZeroFunctional(space)
    elif name == 'ind_unit_ball_1':
        func = odl.functionals.IndicatorLpUnitBall(space, 1)
    elif name == 'ind_unit_ball_2':
        func = odl.functionals.IndicatorLpUnitBall(space, 2)
    elif name == 'ind_unit_ball_pi':
        func = odl.functionals.IndicatorLpUnitBall(space, np.pi)
    elif name == 'ind_unit_ball_inf':
        func = odl.functionals.IndicatorLpUnitBall(space, np.inf)
    elif name == 'product':
        left = odl.functionals.L2Norm(space)
        right = odl.functionals.ConstantFunctional(space, 2)
        func = odl.functionals.FunctionalProduct(left, right)
    elif name == 'quotient':
        dividend = odl.functionals.L2Norm(space)
        divisor = odl.functionals.ConstantFunctional(space, 2)
        func = odl.functionals.FunctionalQuotient(dividend, divisor)
    elif name == 'kl':
        func = odl.functionals.KullbackLeibler(space)
    elif name == 'kl_cc':
        func = odl.functionals.KullbackLeibler(space).convex_conj
    elif name == 'kl_cross_ent':
        func = odl.functionals.KullbackLeiblerCrossEntropy(space)
    elif name == 'kl_cc_cross_ent':
        func = odl.functionals.KullbackLeiblerCrossEntropy(space).convex_conj
    elif name == 'huber':
        func = odl.functionals.Huber(space, gamma=0.1)
    elif name == 'groupl1':
        if isinstance(space, odl.ProductSpace):
            pytest.skip("The `GroupL1Norm` is not supported on `ProductSpace`")
        space = odl.ProductSpace(space, 3)
        func = odl.functionals.GroupL1Norm(space)
    elif name == 'bregman_l2squared':
        point = noise_element(space)
        l2_squared = odl.functionals.L2NormSquared(space)
        subgrad = l2_squared.gradient(point)
        func = odl.functionals.BregmanDistance(l2_squared, point, subgrad)
    elif name == 'bregman_l1':
        point = noise_element(space)
        l1 = odl.functionals.L1Norm(space)
        subgrad = l1.gradient(point)
        func = odl.functionals.BregmanDistance(l1, point, subgrad)
    elif name == 'indicator_simplex':
        diameter = 1.23
        func = odl.functionals.IndicatorSimplex(space, diameter)
    elif name == 'indicator_sum_constraint':
        sum_value = 1.23
        func = odl.functionals.IndicatorSumConstraint(space, sum_value)
    else:
        assert False

    return func


# --- functional tests --- #


def test_derivative(functional):
    """Test for the derivative of a functional.

    The test checks that the directional derivative in a point is the same as
    the inner product of the gradient and the direction, if the gradient is
    defined.
    """
    if isinstance(functional, FUNCTIONALS_WITHOUT_DERIVATIVE):
        # IndicatorFunction has no derivative
        with pytest.raises(NotImplementedError):
            functional.derivative(functional.domain.zero())
        return

    x = noise_element(functional.domain)
    y = noise_element(functional.domain)

    if (isinstance(functional, odl.functionals.KullbackLeibler) or
            isinstance(functional, odl.functionals.KullbackLeiblerCrossEntropy)):
        # The functional is not defined for values <= 0
        x = odl.abs(x)
        y = odl.abs(y)

    if isinstance(functional, KullbackLeiblerConvexConj):
        # The functional is not defined for values >= 1
        x = x - odl.max(x) + 0.99
        y = y - odl.max(y) + 0.99

    # Compute a "small" step size according to dtype of space
    step = float(np.sqrt(_numerical_epsilon(functional.domain.dtype)))

    # Numerical test of gradient, only low accuracy can be guaranteed.
    assert all_almost_equal((functional(x + step * y) - functional(x)) / step,
                            y.inner(functional.gradient(x)),
                            ndigits=1)

    # Check that derivative and gradient is consistent
    assert all_almost_equal(functional.derivative(x)(y),
                            y.inner(functional.gradient(x)))


def test_arithmetic():
    """Test that all standard arithmetic works."""
    space = odl.rn(3)

    # Create elements needed for later
    functional = odl.functionals.L2Norm(space).translated([1, 2, 3])
    functional2 = odl.functionals.L2NormSquared(space)
    operator = odl.IdentityOperator(space) - space.element([4, 5, 6])
    x = noise_element(functional.domain)
    y = noise_element(functional.domain)
    scalar = np.pi

    # Simple tests here, more in depth comes later
    assert functional(x) == functional(x)
    assert functional(x) != functional2(x)
    assert (scalar * functional)(x) == scalar * functional(x)
    assert (scalar * (scalar * functional))(x) == scalar**2 * functional(x)
    assert (functional * scalar)(x) == functional(scalar * x)
    assert ((functional * scalar) * scalar)(x) == functional(scalar**2 * x)
    assert (functional + functional2)(x) == functional(x) + functional2(x)
    assert (functional - functional2)(x) == functional(x) - functional2(x)
    assert (functional * operator)(x) == functional(operator(x))
    assert all_almost_equal((y @ functional)(x), y * functional(x))
    assert all_almost_equal((y @ (y @ functional))(x), (y * y) * functional(x))
    assert all_almost_equal((functional @ y)(x), functional(y * x))
    assert all_almost_equal(((functional @ y) * y)(x), functional((y * y) * x))


def test_left_scalar_mult(space, scalar):
    """Test for right and left multiplication of a functional with a scalar."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    x = noise_element(space)
    func = odl.functionals.L2Norm(space)
    lmul_func = scalar * func

    if scalar == 0:
        assert isinstance(scalar * func, odl.functionals.ZeroFunctional)
        return

    # Test functional evaluation
    assert lmul_func(x) == pytest.approx(scalar * func(x), rel=rtol)

    # Test gradient of left scalar multiplication
    assert all_almost_equal(lmul_func.gradient(x), scalar * func.gradient(x),
                            ndigits)

    # Test derivative of left scalar multiplication
    p = noise_element(space)
    assert all_almost_equal(lmul_func.derivative(x)(p),
                            scalar * (func.derivative(x))(p),
                            ndigits)

    # Test convex conjugate. This requires positive scaling to work
    pos_scalar = abs(scalar)
    neg_scalar = -pos_scalar

    with pytest.raises(ValueError):
        (neg_scalar * func).convex_conj

    assert all_almost_equal((pos_scalar * func).convex_conj(x),
                            pos_scalar * func.convex_conj(x / pos_scalar),
                            ndigits)

    # Test proximal operator. This requires scaling to be positive.
    sigma = 1.2
    with pytest.raises(ValueError):
        (neg_scalar * func).proximal(sigma)

    assert all_almost_equal((pos_scalar * func).proximal(sigma)(x),
                            func.proximal(sigma * pos_scalar)(x))


def test_right_scalar_mult(space, scalar):
    """Test for right and left multiplication of a functional with a scalar."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    x = noise_element(space)
    func = odl.functionals.L2NormSquared(space)
    rmul_func = func * scalar

    if scalar == 0:
        # expecting the constant functional x -> func(0)
        assert isinstance(rmul_func, odl.functionals.ConstantFunctional)
        assert all_almost_equal(rmul_func(x), func(space.zero()),
                                ndigits)
        # Nothing more to do, rest is part of ConstantFunctional test
        return

    # Test functional evaluation
    assert rmul_func(x) == pytest.approx(func(scalar * x), rel=rtol)

    # Test gradient of right scalar multiplication
    assert all_almost_equal(rmul_func.gradient(x),
                            scalar * func.gradient(scalar * x),
                            ndigits)

    # Test derivative of right scalar multiplication
    p = noise_element(space)
    assert all_almost_equal(rmul_func.derivative(x)(p),
                            scalar * func.derivative(scalar * x)(p),
                            ndigits)

    # Test convex conjugate conjugate
    assert all_almost_equal(rmul_func.convex_conj(x),
                            func.convex_conj(x / scalar),
                            ndigits)

    # Test proximal operator
    sigma = 1.2
    assert all_almost_equal(
        rmul_func.proximal(sigma)(x),
        (1.0 / scalar) * func.proximal(sigma * scalar ** 2)(x * scalar),
        ndigits)

    # Verify that for linear functionals, left multiplication is used.
    func = odl.functionals.ZeroFunctional(space)
    assert isinstance(func * scalar, odl.functionals.FunctionalLeftScalarMult)


def test_functional_composition(space):
    """Test composition from the right with an operator."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    func = odl.functionals.L2NormSquared(space)

    # Verify that an error is raised if an invalid operator is used
    # (e.g. wrong range)
    scalar = 2.1
    wrong_space = odl.uniform_discr(1, 2, 10)
    op_wrong = odl.core.operator.ScalingOperator(wrong_space, scalar)

    with pytest.raises(OpTypeError):
        func * op_wrong

    # Test composition with operator from the right
    op = odl.core.operator.ScalingOperator(space, scalar)
    func_op_comp = func * op
    assert isinstance(func_op_comp, odl.functionals.Functional)

    x = noise_element(space)
    assert func_op_comp(x) == pytest.approx(func(op(x)), rel=rtol)

    # Test gradient and derivative with composition from the right
    assert all_almost_equal(func_op_comp.gradient(x),
                            (op.adjoint * func.gradient * op)(x),
                            ndigits)

    p = noise_element(space)
    assert all_almost_equal(func_op_comp.derivative(x)(p),
                            (op.adjoint * func.gradient * op)(x).inner(p),
                            ndigits)


def test_functional_sum(space):
    """Test for the sum of two functionals."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    func1 = odl.functionals.L2NormSquared(space)
    func2 = odl.functionals.L2Norm(space)

    # Verify that an error is raised if one operand is "wrong"
    op = odl.core.operator.IdentityOperator(space)
    with pytest.raises(OpTypeError):
        func1 + op

    wrong_space = odl.uniform_discr(1, 2, 10)
    func_wrong_domain = odl.functionals.L2Norm(wrong_space)
    with pytest.raises(OpTypeError):
        func1 + func_wrong_domain

    func_sum = func1 + func2
    x = noise_element(space)
    p = noise_element(space)

    # Test functional evaluation
    assert func_sum(x) == pytest.approx(func1(x) + func2(x), rel=rtol)

    # Test gradient and derivative
    assert all_almost_equal(func_sum.gradient(x),
                            func1.gradient(x) + func2.gradient(x),
                            ndigits)

    assert (
        func_sum.derivative(x)(p) ==
        pytest.approx(
            func1.gradient(x).inner(p) + func2.gradient(x).inner(p),
            rel=rtol)
    )

    # Verify that proximal raises
    with pytest.raises(NotImplementedError):
        func_sum.proximal

    # Test the convex conjugate raises
    with pytest.raises(NotImplementedError):
        func_sum.convex_conj(x)


def test_functional_plus_scalar(space):
    """Test for sum of functioanl and scalar."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    func = odl.functionals.L2NormSquared(space)
    scalar = -1.3

    # Test for scalar not in the field (field of unifor_discr is RealNumbers)
    complex_scalar = 1j
    with pytest.raises(TypeError):
        func + complex_scalar

    func_scalar_sum = func + scalar
    x = noise_element(space)
    p = noise_element(space)

    # Test for evaluation
    assert func_scalar_sum(x) == pytest.approx(func(x) + scalar, rel=rtol)

    # Test for derivative and gradient
    assert all_almost_equal(func_scalar_sum.gradient(x), func.gradient(x),
                            ndigits)

    assert (
        func_scalar_sum.derivative(x)(p) ==
        pytest.approx(func.gradient(x).inner(p), rel=rtol)
    )

    # Test proximal operator
    sigma = 1.2
    assert all_almost_equal(func_scalar_sum.proximal(sigma)(x),
                            func.proximal(sigma)(x),
                            ndigits)

    # Test convex conjugate
    assert (
        func_scalar_sum.convex_conj(x) ==
        pytest.approx(func.convex_conj(x) - scalar, rel=rtol)
    )

    assert all_almost_equal(func_scalar_sum.convex_conj.gradient(x),
                            func.convex_conj.gradient(x),
                            ndigits)


def test_translation_of_functional(space):
    """Test for the translation of a functional: (f(. - y))^*."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    # The translation; an element in the domain
    translation = noise_element(space)

    test_functional = odl.functionals.L2NormSquared(space)
    translated_functional = test_functional.translated(translation)
    x = noise_element(space)

    # Test for evaluation of the functional
    expected_result = test_functional(x - translation)
    assert all_almost_equal(translated_functional(x), expected_result,
                            ndigits)

    # Test for the gradient
    expected_result = test_functional.gradient(x - translation)
    translated_gradient = translated_functional.gradient
    assert all_almost_equal(translated_gradient(x), expected_result,
                            ndigits)

    # Test for proximal
    sigma = 1.2
    # The helper function below is tested explicitly in proximal_utils_test
    expected_result = odl.solvers.proximal_translation(
        test_functional.proximal, translation)(sigma)(x)
    assert all_almost_equal(translated_functional.proximal(sigma)(x),
                            expected_result, ndigits)

    # Test for conjugate functional
    # The helper function below is tested explicitly further down in this file
    expected_result = odl.functionals.FunctionalQuadraticPerturb(
        test_functional.convex_conj, linear_term=translation)(x)
    assert all_almost_equal(translated_functional.convex_conj(x),
                            expected_result, ndigits)

    # Test for derivative in direction p
    p = noise_element(space)

    # Explicit computation in point x, in direction p: <x/2 + translation, p>
    expected_result = p.inner(test_functional.gradient(x - translation))
    assert all_almost_equal(translated_functional.derivative(x)(p),
                            expected_result, ndigits)

    # Test for optimized implementation, when translating a translated
    # functional
    second_translation = noise_element(space)
    double_translated_functional = translated_functional.translated(
        second_translation)

    # Evaluation
    assert (
        double_translated_functional(x) ==
        pytest.approx(test_functional(x - translation - second_translation),
                      rel=rtol)
    )


# def test_translation_proximal_stepsizes():
#     """Test for stepsize types for proximal of a translated functional."""
#     # Set up space, functional and a point where to evaluate the proximal.
#     space = odl.rn(2)
#     functional = odl.functionals.L2NormSquared(space)
#     translation = functional.translated([0.5, 0.5])
#     x = space.one()

#     # Define different forms of the same stepsize.
#     stepsize = space.element([0.5, 2.0])
#     stepsize_list = [0.5, 2.0]
#     stepsize_array = np.asarray([0.5, 2.0])

#     # Calculate the proximals for each of the stepsizes.
#     y = translation.convex_conj.proximal(stepsize)(x)
#     y_list = translation.convex_conj.proximal(stepsize_list)(x)
#     y_array = translation.convex_conj.proximal(stepsize_array)(x)
#     expected_result = [0.6, 0.0]

#     # Now, all the results should be equal to the expected result.
#     assert all_almost_equal(y, expected_result)
#     assert all_almost_equal(y_list, expected_result)
#     assert all_almost_equal(y_array, expected_result)


def test_multiplication_with_vector(space):
    """Test for multiplying a functional with a vector, both left and right."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    x = noise_element(space)
    y = noise_element(space)
    func = odl.functionals.L2NormSquared(space)

    wrong_space = odl.uniform_discr(1, 2, 10)
    y_other_space = noise_element(wrong_space)

    # Multiplication from the right. Make sure it is a
    # FunctionalRightVectorMult
    func_times_y = func * y
    assert isinstance(func_times_y, odl.functionals.FunctionalRightVectorMult)

    expected_result = func(y * x)
    assert func_times_y(x) == pytest.approx(expected_result, rel=rtol)

    # Test for the gradient.
    # Explicit calculations: 2*y*y*x
    expected_result = 2.0 * y * y * x
    assert all_almost_equal(func_times_y.gradient(x), expected_result,
                            ndigits)

    # Test for convex_conj
    cc_func_times_y = func_times_y.convex_conj
    # Explicit calculations: 1/4 * ||x/y||_2^2
    expected_result = 1.0 / 4.0 * (x / y).norm()**2
    assert cc_func_times_y(x) == pytest.approx(expected_result, rel=rtol)

    # Make sure that right muliplication is not allowed with vector from
    # another space
    with pytest.raises(TypeError):
        func * y_other_space

    # Multiplication from the left. Make sure it is a FunctionalLeftVectorMult
    y_times_func = y @ func
    assert isinstance(y_times_func, odl.FunctionalLeftVectorMult)

    expected_result = y * func(x)
    assert all_almost_equal(y_times_func(x), expected_result, ndigits)

    # Now, multiplication with vector from another space is ok (since it is the
    # same as scaling that vector with the scalar returned by the functional).
    y_other_times_func = y_other_space @ func
    assert isinstance(y_other_times_func, odl.FunctionalLeftVectorMult)

    expected_result = y_other_space * func(x)
    assert all_almost_equal(y_other_times_func(x), expected_result,
                            ndigits)


# Fixtures for test_functional_quadratic_perturb
linear_term = simple_fixture('linear_term', [False, True])
quadratic_coeff = simple_fixture('quadratic_coeff', [0.0, 2.13])


def test_functional_quadratic_perturb(space, linear_term, quadratic_coeff):
    """Test for the functional f(.) + a | . |^2 + <y, .>."""
    # Less strict checking for single precision
    ndigits = dtype_ndigits(space.dtype)
    rtol = dtype_tol(space.dtype)

    orig_func = odl.functionals.L2NormSquared(space)

    if linear_term:
        linear_term_arg = None
        linear_term = space.zero()
    else:
        linear_term_arg = linear_term = noise_element(space)

    # Creating the functional ||x||_2^2 and add the quadratic perturbation
    functional = odl.functionals.FunctionalQuadraticPerturb(
        orig_func,
        quadratic_coeff=quadratic_coeff,
        linear_term=linear_term_arg)

    # Create an element in the space, in which to evaluate
    x = noise_element(space)

    # Test for evaluation of the functional
    assert (
        functional(x) ==
        pytest.approx(orig_func(x) +
                      quadratic_coeff * x.inner(x) +
                      x.inner(linear_term),
                      rel=rtol)
    )

    # Test for the gradient
    assert all_almost_equal(
        functional.gradient(x),
        orig_func.gradient(x) + 2.0 * quadratic_coeff * x + linear_term,
        ndigits
    )

    # Test for the proximal operator if it exists
    sigma = 1.2
    # Explicit computation gives
    c = 1 / np.sqrt(2 * sigma * quadratic_coeff + 1)
    prox = orig_func.proximal(sigma * c ** 2)
    expected_result = prox((x - sigma * linear_term) * c ** 2)
    assert all_almost_equal(functional.proximal(sigma)(x),
                            expected_result,
                            ndigits)

    # Test convex conjugate functional
    if quadratic_coeff == 0:
        expected = orig_func.convex_conj.translated(linear_term)(x)
        assert functional.convex_conj(x) == pytest.approx(expected, rel=rtol)

    # Test proximal of the convex conjugate
    cconj_prox = odl.solvers.proximal_convex_conj(functional.proximal)
    assert all_almost_equal(
        functional.convex_conj.proximal(sigma)(x),
        cconj_prox(sigma)(x),
        ndigits)


def test_bregman(functional):
    """Test for the Bregman distance of a functional."""
    rtol = dtype_tol(functional.domain.dtype)

    if isinstance(functional, FUNCTIONALS_WITHOUT_DERIVATIVE):
        # IndicatorFunction has no gradient
        with pytest.raises(NotImplementedError):
            functional.gradient(functional.domain.zero())
        return

    y = noise_element(functional.domain)
    x = noise_element(functional.domain)

    if (isinstance(functional, odl.functionals.KullbackLeibler) or
            isinstance(functional, odl.functionals.KullbackLeiblerCrossEntropy)):
        # The functional is not defined for values <= 0
        x = odl.abs(x)
        y = odl.abs(y)

    if isinstance(functional, KullbackLeiblerConvexConj):
        # The functional is not defined for values >= 1
        x = x - odl.max(x) + 0.99
        y = y - odl.max(y) + 0.99

    grad = functional.gradient(y)
    quadratic_func = odl.functionals.QuadraticForm(
        vector=-grad, constant=-functional(y) + grad.inner(y))
    expected_func = functional + quadratic_func

    assert (
        functional.bregman(y, grad)(x) ==
        pytest.approx(expected_func(x), rel=rtol)
    )


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for `diff_ops`."""

from __future__ import division

import pytest

import odl
from odl.core.discr.diff_ops import (
    Divergence, Gradient, Laplacian, PartialDerivative, finite_diff)
from odl.core.util.testutils import (
    all_almost_equal, all_equal, dtype_tol, noise_element, simple_fixture)
from odl.core.array_API_support import get_array_and_backend, odl_all_equal

# --- pytest fixtures --- #


method = simple_fixture('method', ['central', 'forward', 'backward'])
padding = simple_fixture('padding', [('constant', 0), ('constant', 1),
                                     'symmetric', 'periodic',
                                     'order0', 'order1', 'order2'])


@pytest.fixture(scope="module", params=[1, 2, 3], ids=['1d', '2d', '3d'])
def space(request, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    ndim = request.param

    return odl.uniform_discr([0] * ndim, [1] * ndim, [5] * ndim, impl=impl, device=device)

@pytest.fixture(scope="module")
def data(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    return odl.uniform_discr(0, 1, 10, impl=impl, device=device).element([0.5, 1, 3.5, 2, -.5, 3, -1, -1, 0, 3])
# --- finite_diff --- #


def test_finite_diff_invalid_args(data):
    """Test finite difference function for invalid arguments."""
    arr, backend = get_array_and_backend(data)
    # Test that old "edge order" argument fails.
    with pytest.raises(TypeError):
        finite_diff(data, axis=0, edge_order=0)

    # at least a two-element array is required
    with pytest.raises(ValueError):
        finite_diff(backend.array_constructor([0.0]), axis=0)

    # axis
    with pytest.raises(IndexError):
        finite_diff(data, axis=2)

    # in-place argument
    # size is an attribute of numpy arrays but of method of pytorch tensors
    out = backend.array_namespace.zeros(len(arr) + 1)
    with pytest.raises(ValueError):
        finite_diff(data, axis=0, out=out)
    with pytest.raises(ValueError):
        finite_diff(data, axis=0, dx=0)

    # wrong method
    with pytest.raises(ValueError):
        finite_diff(data, axis=0, method='non-method')


def test_finite_diff_explicit(data):
    """Compare finite differences function to explicit computation."""

    # phantom data
    arr, backend = get_array_and_backend(data)
    ns = backend.array_namespace

    # explicitly calculated finite difference
    diff_ex = ns.zeros_like(arr)

    # interior: second-order accurate differences
    diff_ex[1:-1] = (arr[2:] - arr[:-2]) / 2.0

    # default: out=None, axis=0, dx=1.0, zero_padding=None, method='forward'
    diff = finite_diff(arr, axis=0, dx=1.0, out=None,
                       pad_mode='constant')
    assert odl_all_equal(diff, finite_diff(arr, axis=0))

    # boundary: one-sided second-order accurate forward/backward difference
    diff = finite_diff(arr, axis=0, dx=1.0, out=None,
                       method='central', pad_mode='order2')
    diff_ex[0] = -(3 * arr[0] - 4 * arr[1] + arr[2]) / 2.0
    diff_ex[-1] = (3 * arr[-1] - 4 * arr[-2] + arr[-3]) / 2.0
    assert odl_all_equal(diff, diff_ex)

    # non-unit step length
    dx = 0.5
    diff = finite_diff(arr, axis=0, dx=dx, method='central', out=None,
                       pad_mode='order2')
    assert odl_all_equal(diff, diff_ex / dx)

    # boundary: second-order accurate central differences with zero padding
    diff = finite_diff(arr, axis=0, method='central', pad_mode='constant',
                       pad_const=0)
    diff_ex[0] = arr[1] / 2.0
    diff_ex[-1] = -arr[-2] / 2.0
    assert odl_all_equal(diff, diff_ex)

    # boundary: one-sided first-order forward/backward difference without zero
    # padding
    diff = finite_diff(arr, axis=0, method='central', pad_mode='order1')
    diff_ex[0] = arr[1] - arr[0]  # 1st-order accurate forward difference
    diff_ex[-1] = arr[-1] - arr[-2]  # 1st-order accurate backward diff.
    assert odl_all_equal(diff, diff_ex)

    # different edge order really differ
    df1 = finite_diff(arr, axis=0, method='central', pad_mode='order1')
    df2 = finite_diff(arr, axis=0, method='central', pad_mode='order2')
    assert odl_all_equal(df1[1:-1], diff_ex[1:-1])
    assert odl_all_equal(df2[1:-1], diff_ex[1:-1])
    assert df1[0] != df2[0]
    assert df1[-1] != df2[-1]

    # in-place evaluation
    out = ns.zeros_like(arr)
    assert out is finite_diff(arr, axis=0, out=out)
    assert odl_all_equal(out, finite_diff(arr, axis=0))
    assert out is not finite_diff(arr, axis=0)

    # axis
    arr = backend.array_constructor([[0., 2., 4., 6., 8.],
                    [1., 3., 5., 7., 9.]])
    df0 = finite_diff(arr, axis=0, pad_mode='order1')
    darr0 = 1 * ns.ones(arr.shape)
    assert odl_all_equal(df0, darr0)
    darr1 = 2 * ns.ones(arr.shape)
    df1 = finite_diff(arr, axis=1, pad_mode='order1')
    assert odl_all_equal(df1, darr1)

    # complex arrays
    arr = backend.array_constructor([0., 1., 2., 3., 4.]) + 1j * backend.array_constructor([10., 9., 8., 7.,
                                                          6.])
    diff = finite_diff(arr, axis=0, pad_mode='order1')
    assert all(diff.real == 1)
    assert all(diff.imag == -1)


def test_finite_diff_symmetric_padding(data):
    """Finite difference using replicate padding."""

    # Using replicate padding forward and backward differences have zero
    # derivative at the upper or lower endpoint, respectively
    assert finite_diff(data, axis=0, method='forward',
                       pad_mode='symmetric')[-1] == 0
    assert finite_diff(data, axis=0, method='backward',
                       pad_mode='symmetric')[0] == 0

    diff = finite_diff(data, axis=0, method='central', pad_mode='symmetric')
    assert diff[0] == (data[1] - data[0]) / 2
    assert diff[-1] == (data[-1] - data[-2]) / 2


def test_finite_diff_constant_padding(data):
    """Finite difference using constant padding."""

    for pad_const in [-1, 0, 1]:
        diff_forward = finite_diff(data, axis=0, method='forward',
                                   pad_mode='constant',
                                   pad_const=pad_const)

        assert diff_forward[0] == data[1] - data[0]
        assert diff_forward[-1] == pad_const - data[-1]

        diff_backward = finite_diff(data, axis=0, method='backward',
                                    pad_mode='constant',
                                    pad_const=pad_const)

        assert diff_backward[0] == data[0] - pad_const
        assert diff_backward[-1] == data[-1] - data[-2]

        diff_central = finite_diff(data, axis=0, method='central',
                                   pad_mode='constant',
                                   pad_const=pad_const)

        assert diff_central[0] == (data[1] - pad_const) / 2
        assert diff_central[-1] == (pad_const - data[-2]) / 2


def test_finite_diff_periodic_padding(data):
    """Finite difference using periodic padding."""

    diff_forward = finite_diff(data, axis=0, method='forward',
                               pad_mode='periodic')

    assert diff_forward[0] == data[1] - data[0]
    assert diff_forward[-1] == data[0] - data[-1]

    diff_backward = finite_diff(data, axis=0, method='backward',
                                pad_mode='periodic')

    assert diff_backward[0] == data[0] - data[-1]
    assert diff_backward[-1] == data[-1] - data[-2]

    diff_central = finite_diff(data, axis=0, method='central',
                               pad_mode='periodic')

    assert diff_central[0] == (data[1] - data[-1]) / 2
    assert diff_central[-1] == (data[0] - data[-2]) / 2


# --- PartialDerivative --- #


def test_part_deriv_init():
    """Check initialization of ``PartialDerivative``."""
    space = odl.uniform_discr([0, 0], [1, 1], (4, 5))

    op = PartialDerivative(space, axis=0)
    assert repr(op) != ''
    op = PartialDerivative(space, axis=1)
    assert repr(op) != ''
    op = PartialDerivative(space, axis=0, range=space.astype('float32'))
    assert repr(op) != ''
    op = PartialDerivative(space, axis=0, method='central')
    assert repr(op) != ''
    op = PartialDerivative(space, axis=0, pad_const=1)
    assert repr(op) != ''
    op = PartialDerivative(space, axis=0, pad_mode='order1')
    assert repr(op) != ''

    with pytest.raises(TypeError):
        PartialDerivative(odl.rn(1))


def test_part_deriv(space, method, padding):
    """Discretized partial derivative."""
    if isinstance(padding, tuple):
        pad_mode, pad_const = padding
    else:
        pad_mode, pad_const = padding, 0

    dom_vec = noise_element(space)
    dom_vec_arr = dom_vec.asarray()

    for axis in range(space.ndim):
        partial = PartialDerivative(space, axis=axis, method=method,
                                    pad_mode=pad_mode,
                                    pad_const=pad_const)

        # Compare to helper function
        dx = space.cell_sides[axis]
        diff = finite_diff(dom_vec_arr, axis=axis, dx=dx, method=method,
                        pad_mode=pad_mode,
                        pad_const=pad_const)

        partial_vec = partial(dom_vec)
        assert all_almost_equal(partial_vec, diff)

        # Test adjoint operator
        derivative = partial.derivative()
        ran_vec = noise_element(space)
        deriv_vec = derivative(dom_vec)
        adj_vec = derivative.adjoint(ran_vec)
        lhs = ran_vec.inner(deriv_vec)
        rhs = dom_vec.inner(adj_vec)

        # Check not to use trivial data
        assert lhs != 0
        assert rhs != 0
        assert lhs == pytest.approx(rhs, rel=dtype_tol(space.dtype))


# --- Gradient --- #


def test_gradient_init(odl_impl_device_pairs):
    """Check initialization of ``Gradient``."""
    impl, device = odl_impl_device_pairs
    space = odl.uniform_discr([0, 0], [1, 1], (4, 5), impl=impl, device=device)
    vspace = space ** 2

    op = Gradient(space)
    assert repr(op) != ''
    op = Gradient(range=vspace)
    assert repr(op) != ''
    op = Gradient(space, range=space.astype('float32') ** 2)
    assert repr(op) != ''
    op = Gradient(space, method='central')
    assert repr(op) != ''
    op = Gradient(space, pad_const=1)
    assert repr(op) != ''
    op = Gradient(space, pad_mode='order1')
    assert repr(op) != ''

    with pytest.raises(TypeError):
        Gradient(odl.rn(1))

    with pytest.raises(TypeError):
        Gradient(space, range=space)

    with pytest.raises(ValueError):
        Gradient(space, range=space ** 3)


def test_gradient(space, method, padding):
    """Discretized spatial gradient operator."""
    with pytest.raises(TypeError):
        Gradient(odl.rn(1), method=method)

    if isinstance(padding, tuple):
        pad_mode, pad_const = padding
    else:
        pad_mode, pad_const = padding, 0

    # DiscretizedSpaceElement
    dom_vec = noise_element(space)
    dom_vec_arr = dom_vec.asarray()

    # gradient
    grad = Gradient(space, method=method,
                    pad_mode=pad_mode,
                    pad_const=pad_const)
    grad_vec = grad(dom_vec)
    assert len(grad_vec) == space.ndim

    # computation of gradient components with helper function
    for axis, dx in enumerate(space.cell_sides):
        diff = finite_diff(dom_vec_arr, axis=axis, dx=dx, method=method,
                        pad_mode=pad_mode,
                        pad_const=pad_const)

        assert all_almost_equal(grad_vec[axis].asarray(), diff)

    # Test adjoint operator
    derivative = grad.derivative()
    ran_vec = noise_element(derivative.range)
    deriv_grad_vec = derivative(dom_vec)
    adj_grad_vec = derivative.adjoint(ran_vec)
    lhs = ran_vec.inner(deriv_grad_vec)
    rhs = dom_vec.inner(adj_grad_vec)

    # Check not to use trivial data
    assert lhs != 0
    assert rhs != 0
    assert lhs == pytest.approx(rhs, rel=dtype_tol(space.dtype))

    # Higher-dimensional arrays
    lin_size = 3
    for ndim in [1, 3, 6]:
        space = odl.uniform_discr([0.] * ndim, [1.] * ndim, [lin_size] * ndim)
        dom_vec = odl.core.phantom.cuboid(space, [0.2] * ndim, [0.8] * ndim)

        grad = Gradient(space, method=method, pad_mode=pad_mode,
                        pad_const=pad_const)
        grad(dom_vec)

# --- Divergence --- #


def test_divergence_init(odl_impl_device_pairs):
    """Check initialization of ``Divergence``."""
    impl, device = odl_impl_device_pairs
    space = odl.uniform_discr([0, 0], [1, 1], (4, 5), impl=impl, device=device)
    vspace = space ** 2

    op = Divergence(vspace)
    assert repr(op) != ''
    op = Divergence(range=space)
    assert repr(op) != ''
    op = Divergence(vspace, range=space.astype('float32'))
    assert repr(op) != ''
    op = Divergence(vspace, method='central')
    assert repr(op) != ''
    op = Divergence(vspace, pad_const=1)
    assert repr(op) != ''
    op = Divergence(vspace, pad_mode='order1')
    assert repr(op) != ''

    with pytest.raises(TypeError):
        Divergence(odl.rn(1) ** 2)

    with pytest.raises(TypeError):
        Divergence(vspace, range=vspace)

    with pytest.raises(ValueError):
        Divergence(space ** 3, range=space)


def test_divergence(space, method, padding):
    """Discretized spatial divergence operator."""
    # Invalid space
    with pytest.raises(TypeError):
        Divergence(range=odl.rn(1), method=method)

    if isinstance(padding, tuple):
        pad_mode, pad_const = padding
    else:
        pad_mode, pad_const = padding, 0

    # Operator instance
    div = Divergence(range=space, method=method,
                    pad_mode=pad_mode,
                    pad_const=pad_const)

    # Apply operator
    dom_vec = noise_element(div.domain)
    div_dom_vec = div(dom_vec)

    # computation of divergence with helper function
    expected_result = space.array_namespace.zeros(space.shape, dtype=space.dtype, device=space.device)
    for axis, dx in enumerate(space.cell_sides):
        expected_result += finite_diff(dom_vec[axis], axis=axis, dx=dx,
                                    method=method, pad_mode=pad_mode,
                                    pad_const=pad_const)

    assert all_almost_equal(expected_result, div_dom_vec.asarray())

    # Adjoint operator
    derivative = div.derivative()
    deriv_div_dom_vec = derivative(dom_vec)
    ran_vec = noise_element(div.range)
    adj_div_ran_vec = derivative.adjoint(ran_vec)

    # Adjoint condition
    lhs = ran_vec.inner(deriv_div_dom_vec)
    rhs = dom_vec.inner(adj_div_ran_vec)
    # Check not to use trivial data
    assert lhs != 0
    assert rhs != 0
    assert lhs == pytest.approx(rhs, rel=dtype_tol(space.dtype))


# --- Laplacian --- #

def test_laplacian_init(odl_impl_device_pairs):
    """Check initialization of ``Laplacian``."""
    impl, device = odl_impl_device_pairs
    space = odl.uniform_discr([0, 0], [1, 1], (4, 5), impl=impl, device=device)

    op = Laplacian(space)
    assert repr(op) != ''
    op = Laplacian(space, range=space.astype('float32'))
    assert repr(op) != ''
    op = Laplacian(space, pad_const=1)
    assert repr(op) != ''
    op = Laplacian(space, pad_mode='order0')
    assert repr(op) != ''

    with pytest.raises(TypeError):
        Laplacian(odl.rn(1))


def test_laplacian(space, padding):
    """Discretized spatial laplacian operator."""
    if isinstance(padding, tuple):
        pad_mode, pad_const = padding
    else:
        pad_mode, pad_const = padding, 0

    if pad_mode in ('order1', 'order2'):
        return  # these pad modes not supported for laplacian

    # Operator instance
    lap = Laplacian(space, pad_mode=pad_mode, pad_const=pad_const)

    # Apply operator
    dom_vec = noise_element(space)
    div_dom_vec = lap(dom_vec)

    # computation of divergence with helper function
    expected_result = space.array_namespace.zeros(space.shape, device=space.device, dtype=space.dtype)
    for axis, dx in enumerate(space.cell_sides):
        diff_f = finite_diff(dom_vec.asarray(), axis=axis, dx=dx ** 2,
                            method='forward', pad_mode=pad_mode,
                            pad_const=pad_const)
        diff_b = finite_diff(dom_vec.asarray(), axis=axis, dx=dx ** 2,
                            method='backward', pad_mode=pad_mode,
                            pad_const=pad_const)
        expected_result += diff_f - diff_b

    assert all_almost_equal(expected_result, div_dom_vec.asarray())

    # Adjoint operator
    derivative = lap.derivative()
    deriv_lap_dom_vec = derivative(dom_vec)
    ran_vec = noise_element(lap.range)
    adj_lap_ran_vec = derivative.adjoint(ran_vec)

    # Adjoint condition
    lhs = ran_vec.inner(deriv_lap_dom_vec)
    rhs = dom_vec.inner(adj_lap_ran_vec)

    # Check not to use trivial data
    assert lhs != 0
    assert rhs != 0
    assert lhs == pytest.approx(rhs, rel=dtype_tol(space.dtype))


if __name__ == '__main__':
    odl.core.util.test_file(__file__)
